#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" Command-line tool to merge images of external file-formats
into the master NeXus file
"""

import sys
import os
import re
import shutil
import fabio
import signal
import argparse
import numpy
import json

from .filenamegenerator import FilenameGenerator
from .nxsargparser import (Runner, NXSArgParser, ErrorException)
from . import filewriter


if sys.version_info > (3,):
    unicode = str
    long = int
else:
    bytes = str


WRITERS = {}
try:
    from . import h5pywriter
    WRITERS["h5py"] = h5pywriter
except Exception:
    pass

try:
    from . import h5cppwriter
    WRITERS["h5cpp"] = h5cppwriter
except Exception:
    pass


pTc = {
    "long": long,
    "str": str,
    "unicode": unicode,
    "bool": bool,
    "int": int,
    "int64": int,
    "int32": int,
    "int16": int,
    "int8": int,
    "uint": int,
    "uint64": int,
    "uint32": int,
    "uint16": int,
    "uint8": int,
    "float": float,
    "float64": float,
    "float32": float,
    "string": str,
}


def _tostr(text):
    """ converts text  to str type

    :param text: text
    :type text: :obj:`bytes` or :obj:`unicode`
    :returns: text in str type
    :rtype: :obj:`str`
    """
    if isinstance(text, str):
        return text
    elif sys.version_info > (3,):
        return str(text, "utf8")
    else:
        return str(text)


def filegenerator(filestr, pattern=None):
    """ provides file name generator from file string

    :param filestr: file string
    :type: filestr: :obj:`str`
    :returns: file name generator or a list of file names
    :rtype: :class:`methodinstance`
    """
    if pattern is None:
        pattern = re.compile(".+:\\d+:\\d+")
    if pattern.match(filestr):
        return FilenameGenerator.from_slice(filestr)
    else:
        def _files():
            return [filestr]
        return _files


def splitcoords(crdstr):
    """ splits coordinate string

    :param crdstr: cordinate string
    :type crdstr: cordinate string
    :returns: a list ofr coordinates in tuples
    :rtype: :obj:`list` <:obj:`tuple` < :obj:`int` >>
    """
    if crdstr is None:
        return []
    crds = []
    scrds = crdstr.replace(';', ' ').replace(':', ' ').split(" ")
    for crd in scrds:
        if crd.startswith("[") and crd.endswith("]"):
            crd = crd[1:-1]
        scds = crd.split(",")
        nc = []
        for cd in scds:
            nc.append(crdtoint(cd.strip()))
        crds.append(tuple(nc))
    return crds


def crdtoint(crd):
    """ convert coorinate to int or  None or Unlimited

    :param crd: cordinate as string
    :type crd: cordinate as string
    :returns: converted coordinate
    :rtype: :obj:`int`
    """
    if crd in ["u", "U"]:
        crd = filewriter.unlimited()
    elif crd in ["", "None", "null"]:
        crd = None
    else:
        try:
            crd = int(crd)
        except Exception as e:
            print(str(e))
            sys.stderr.write(
                "Warning: wrong coordinate %s. Converted to None\n")
            crd = None
    return crd


def splitslices(crdstr):
    """ splits coordinate string

    :param crdstr: cordinate string
    :type crdstr: cordinate string
    :returns: a list ofr coordinates in tuples
    :rtype: :obj:`list` <:obj:`tuple` < :obj:`int` >>
    """
    crds = []
    if crdstr is None:
        return []
    scrds = crdstr.replace(';', ' ').split(" ")
    for crd in scrds:
        scds = crd.split(",")
        nc = []
        for cd in scds:
            start = None
            stop = None
            step = None
            sls = cd.split(":")
            if len(sls) == 1:
                stop = crdtoint(sls[0])
            if len(sls) == 2:
                start = crdtoint(sls[0])
                stop = crdtoint(sls[1])
            if len(sls) > 2:
                start = crdtoint(sls[0])
                stop = crdtoint(sls[1])
                step = crdtoint(sls[2])

            nc.append(slice(start, stop, step))
        crds.append(tuple(nc))
    return crds


def getcompression(compression):
    """ converts compression string to a deflate level parameter
        or list with [filterid, opt1, opt2, ...]

    :param compression: compression string
    :type compression: :obj:`str`
    :returns: deflate level parameter
              or list with [filterid, opt1, opt2, ...]
    :rtype: :obj:`int` or :obj:`list` < :obj:`int` > or `None`

    """
    if compression:
        if isinstance(compression, int) or ":" not in compression:
            level = None
            try:
                level = int(compression)
            except Exception:
                raise Exception(
                    "Error: argument -c/--compression: "
                    "invalid int value: '%s'\n" % compression)
            return level
        else:
            opts = None
            try:
                sfid, sopts = compression.split(":")
                opts = [int(sfid)]
                opts.extend([int(opt) for opt in sopts.split(",")])
            except Exception:
                raise Exception(
                    "Error: argument -c/--compression: "
                    "invalid format: '%s'\n" % compression)
            return opts
        return


class Linker(object):

    """ Create external and internal links of NeXus files
    """

    def __init__(self, nexusfilepath, target, name=None,
                 storeold=False, testmode=False, writer=None):
        """ The constructor creates the collector object

        :param nexusfilepath: the nexus file name and nexus path
        :type nexusfilepath: :obj:`str`
        :param target: the nexus file name and nexus path
        :type target: :obj:`str`
        :param storeold: if backup the input file
        :type storeold: :obj:`bool`
        :param testmode: if run in a test mode
        :type testmode: :obj:`bool`
        :param writer: the writer module
        :type writer: :obj:`str`
        """
        self.__target = target
        self.__name = name
        if not name:
            self.__name = target.split("/")[-1]
        self.__testmode = testmode
        self.__storeold = storeold
        self.__wrmodule = None
        self.__nexuspath = None
        self.__nexusfilename, self.__nexuspath = \
            nexusfilepath.split(":/")

        if writer and writer.lower() in WRITERS.keys():
            self.__wrmodule = WRITERS[writer.lower()]
        self.__siginfo = dict(
            (signal.__dict__[sname], sname)
            for sname in ('SIGINT', 'SIGHUP', 'SIGALRM', 'SIGTERM'))

        for sig in self.__siginfo.keys():
            signal.signal(sig, self._signalhandler)

    def _signalhandler(self, sig, _):
        """ signal handler

        :param sig: signal name, i.e. 'SIGINT', 'SIGHUP', 'SIGALRM', 'SIGTERM'
        :type sig: :obj:`str`
        """
        if sig in self.__siginfo.keys():
            self.__break = True
            print("terminated by %s" % self.__siginfo[sig])

    def _createtmpfile(self):
        """ creates temporary file
        """
        self.__tempfilename = self.__nexusfilename + ".__nxscollect_temp__"
        while os.path.exists(self.__tempfilename):
            self.__tempfilename += "_"
        shutil.copy2(self.__nexusfilename, self.__tempfilename)

    def _storeoldfile(self):
        """ makes back up of the input file
        """
        temp = self.__nexusfilename + ".__nxscollect_old__"
        while os.path.exists(temp):
            temp += "_"
        shutil.move(self.__nexusfilename, temp)

    def link(self):
        """ creates NeXus link
        """
        self._createtmpfile()
        path = self.__nexuspath
        try:
            self.__nxsfile = filewriter.open_file(
                self.__tempfilename, readonly=False,
                writer=self.__wrmodule)
            root = self.__nxsfile.root()
            groups = path.split("/")
            parent = root
            tgr = ""
            for gr in groups:
                if gr:
                    if ":" in gr:
                        gr, tgr = gr.split(":", 1)
                    if parent is not None and gr in parent.names():
                        parent = parent.open(gr)
                    else:
                        if not tgr:
                            tgr = "NX" + gr
                        if not self.__testmode:
                            parent = parent.create_group(gr, tgr)
                        else:
                            parent = None

            if parent:
                print("link: target %s at %s://%s as %s" %
                      (self.__target, self.__nexusfilename,
                       parent.path, self.__name))
            else:
                print("link: target %s at %s as %s" %
                      (self.__target, path, self.__name))
            if not self.__testmode:
                filewriter.link(self.__target, parent, self.__name)

            if self.__storeold:
                self._storeoldfile()
            shutil.move(self.__tempfilename, self.__nexusfilename)
        except Exception as e:
            print(str(e))
            os.remove(self.__tempfilename)


class TargetFieldView(object):

    """ target field map
    """
    def __init__(self, filename, path,
                 shape=None, hyperslab=None, maxshape=None):
        """ constructor

        :param filename: file name
        :type filename: :obj:`str`
        :param path: nexus field path with its name
        :type path: :obj:`str`
        :param shape: field shape
        :type shape:  :obj:`list` <:obj:`int`>
        :param hyperslab: field hyperslab or slices
        :type hyperslab: :class:`filewriter.FTHyperslab`
        :param maxshape: maximal field shape
        :type maxshape:  :obj:`list` <:obj:`int`>
        """

        #: :obj:`str` file name
        self.filename = filename
        #: :obj:`str` nexus field path with its name
        self.path = path
        #: :obj:`list` <:obj:`int`> field shape
        self.shape = shape
        #: :obj:`list` <:obj:`int`> field maximal shape
        self.maxshape = maxshape
        #: :class:`filewriter.FTHyperslab` field hyperslab or slices
        self.hyperslab = hyperslab or filewriter.FTHyperslab()
        #: :obj:`list` <:obj:`slice`>
        self.slices = None


class LayoutField(object):

    def __init__(self, target, hyperslab=None):

        """ constructor

        :param hyperslab: target field object
        :type hyperslab: :class:`TargetFieldView`
        :param hyperslab: field hyperslab or slices
        :type hyperslab: :class:`filewriter.FTHyperslab`
        """

        #: :class:`filewriter.FTHyperslab` layout hyperslab or slices
        self.hyperslab = hyperslab or filewriter.FTHyperslab()
        #: :obj:`list` <:obj:`slice`>
        self.slices = None
        #: :class:`TargetFieldView` target field object
        self.target = target
        #: :obj:`tuple` vds shape in layout
        self.shape = []


class TargetFieldsLayout(list):

    def __init__(self, exfieldpaths='', exfieldshapes='', shapes=None,
                 separator=','):

        """ constructor

        :param exfieldpaths: target field paths
        :type exfieldpaths: :obj:`str`
        :param exfieldshapes: target field shapes
        :type exfieldshapes: :obj:`str`
        :param shapes: target field shapes
        :type shapes: :obj:`list`<:obj:`tuple`<:obj:`int`> >
        :param exfieldpaths: separator of field path strings
        :type exfieldpaths: :obj:`str`
        """

        if separator:
            files = exfieldpaths.split(separator)
        else:
            files = [exfieldpaths]

        for filestr in files:
            inputfiles = filegenerator(filestr)
            for fc in inputfiles():
                fnph = fc.split(":/")
                ph = "/data"
                if len(fnph) > 0:
                    fn = fnph[0]
                    if len(fnph) > 1:
                        ph = fnph[1]
                    efd = TargetFieldView(fn, ph)
                    lfd = LayoutField(efd)
                    self.append(lfd)
        efshapes = splitcoords(exfieldshapes)

        for i, lfd in enumerate(self):
            if i < len(efshapes):
                lfd.target.shape = efshapes[i]
            if i < len(shapes):
                lfd.shape = shapes[i]
                # print(lfd.shape)

    def add_target_hyperslabs(self, offsets, blocks, counts, strides):
        """ add target hyperslabs

        :param offsets: target offsets
        :type offsets: :obj:`str`
        :param blocks: target blocks
        :type blocks: :obj:`str`
        :param counts: target counts
        :type counts: :obj:`str`
        :param strides: target strides
        :type strides: :obj:`str`
        """

        offs = splitcoords(offsets)
        blks = splitcoords(blocks)
        cnts = splitcoords(counts)
        stds = splitcoords(strides)

        for i, lfd in enumerate(self):
            if i < len(offs):
                lfd.target.hyperslab.offset = offs[i]
        for i, lfd in enumerate(self):
            if i < len(blks):
                lfd.target.hyperslab.block = blks[i]
        for i, lfd in enumerate(self):
            if i < len(cnts):
                lfd.target.hyperslab.count = cnts[i]
        for i, lfd in enumerate(self):
            if i < len(stds):
                lfd.target.hyperslab.stride = stds[i]

    def add_layout_hyperslabs(self, offsets, blocks, counts, strides):
        """ add layout hyperslabs

        :param offsets: layout offsets
        :type offsets: :obj:`str`
        :param blocks: layout blocks
        :type blocks: :obj:`str`
        :param counts: layout counts
        :type counts: :obj:`str`
        :param strides: layout strides
        :type strides: :obj:`str`
        """
        offs = splitcoords(offsets)
        blks = splitcoords(blocks)
        cnts = splitcoords(counts)
        stds = splitcoords(strides)

        for i, lfd in enumerate(self):
            if i < len(offs):
                lfd.hyperslab.offset = offs[i]
        for i, lfd in enumerate(self):
            if i < len(blks):
                lfd.hyperslab.block = blks[i]
        for i, lfd in enumerate(self):
            if i < len(cnts):
                lfd.hyperslab.count = cnts[i]
        for i, lfd in enumerate(self):
            if i < len(stds):
                lfd.hyperslab.stride = stds[i]

    def add_layout_slices(self, slices):
        """ add layout slices

        :param slices: layout slices
        :type slices: :obj:`str`
        """
        slices = splitslices(slices)
        for i, lfd in enumerate(self):
            if i < len(slices):
                lfd.slices = slices[i]

    def add_target_slices(self, slices):
        """ add taget slices

        :param slices: target slices
        :type slices: :obj:`str`
        """
        slices = splitslices(slices)
        for i, lfd in enumerate(self):
            if i < len(slices):
                lfd.target.slices = slices[i]


class VirtualDataset(object):

    """ Create virtual dataset in the master NeXus files
    """

    def __init__(self, nexusfilepath, options, writer=None):
        """ The constructor creates the collector object

        :param nexusfilepath: the nexus file name and nexus path
        :type nexusfilepath: :obj:`str`
        :param options: parser options
        :type options: :class:`argparse.Namespace`
        :param writer: the writer module
        :type writer: :obj:`str`
        """
        shape = options.shape
        if shape.startswith("[") and shape.endswith("]"):
            shape = shape[1:-1]
        shape = splitcoords(shape)
        self.__shape = shape[0] if shape else []
        self.__shapes = splitcoords(options.shapes)
        if self.__shape:
            for si, sh in enumerate(self.__shapes):
                changed = False
                lsh = list(sh)
                for di, dm in enumerate(lsh):
                    if dm is None:
                        lsh[di] = self.__shape[di]
                        changed = True
                if changed:
                    self.__shapes[si] = tuple(lsh)
        self.__dtype = options.dtype

        self.__maxshape = None
        self.__fillvalue = options.fillvalue

        self.__storeold = not options.replaceold
        self.__testmode = options.testmode

        self.__wrmodule = None
        self.__nexuspath = None
        self.__nexusfilename, self.__nexuspath = \
            nexusfilepath.split(":/")

        self.__ltfields = TargetFieldsLayout(
            options.targetfields,
            options.targetshapes,
            self.__shapes,
            options.separator
        )
        self.__ltfields.add_target_hyperslabs(
            options.targetoffsets, options.targetblocks,
            options.targetcounts, options.targetstrides)

        if options.targetslices:
            self.__ltfields.add_target_slices(options.targetslices)

        self.__ltfields.add_layout_hyperslabs(
            options.offsets, options.blocks,
            options.counts, options.strides)
        if options.slices:
            self.__ltfields.add_layout_slices(options.slices)

        if writer and writer.lower() in WRITERS.keys():
            self.__wrmodule = WRITERS[writer.lower()]
        self.__siginfo = dict(
            (signal.__dict__[sname], sname)
            for sname in ('SIGINT', 'SIGHUP', 'SIGALRM', 'SIGTERM'))

        for sig in self.__siginfo.keys():
            signal.signal(sig, self._signalhandler)

    def _signalhandler(self, sig, _):
        """ signal handler

        :param sig: signal name, i.e. 'SIGINT', 'SIGHUP', 'SIGALRM', 'SIGTERM'
        :type sig: :obj:`str`
        """
        if sig in self.__siginfo.keys():
            self.__break = True
            print("terminated by %s" % self.__siginfo[sig])

    def _createtmpfile(self):
        """ creates temporary file
        """
        self.__tempfilename = self.__nexusfilename + ".__nxscollect_temp__"
        while os.path.exists(self.__tempfilename):
            self.__tempfilename += "_"
        shutil.copy2(self.__nexusfilename, self.__tempfilename)

    def _storeoldfile(self):
        """ makes back up of the input file
        """
        temp = self.__nexusfilename + ".__nxscollect_old__"
        while os.path.exists(temp):
            temp += "_"
        shutil.move(self.__nexusfilename, temp)

    def create(self):
        """ creates VDS
        """
        self._createtmpfile()
        path = self.__nexuspath
        try:
            self.__nxsfile = filewriter.open_file(
                self.__tempfilename, readonly=False,
                writer=self.__wrmodule)
            root = self.__nxsfile.root()
            groups = path.split("/") or ["data"]
            parent = root
            tgr = ""
            fieldname = groups[-1] or "data"
            for gr in groups[:-1]:
                if gr:
                    if ":" in gr:
                        gr, tgr = gr.split(":", 1)
                    if parent is not None and gr in parent.names():
                        parent = parent.open(gr)
                    else:
                        if not tgr:
                            tgr = "NX" + gr
                        if not self.__testmode:
                            parent = parent.create_group(gr, tgr)
                        else:
                            parent = None
            filewriter.module = self.__wrmodule
            layout = filewriter.virtual_field_layout(
                self.__shape, self.__dtype, self.__maxshape, parent)
            for flm in self.__ltfields:
                efield = filewriter.target_field_view(
                    flm.target.filename, flm.target.path,
                    flm.target.shape or flm.shape,
                    flm.target.maxshape, parent=parent)
                layout.add(flm.hyperslab, efield, flm.target.hyperslab,
                           flm.shape)
                if parent:
                    print("vds: target %s://%s %s at %s/%s" %
                          (flm.target.filename, flm.target.path,
                           flm.target.shape, parent.path, fieldname))
                else:
                    print("vds: target %s://%s %s at %s/%s" %
                          (flm.target.filename, flm.target.path,
                           flm.target.shape, path, fieldname))
            if not self.__testmode:
                fillvalue = pTc[_tostr(self.__dtype)](self.__fillvalue or 0)
                fd = parent.create_virtual_field(fieldname, layout, fillvalue)
                fd.close()

            if self.__storeold:
                self._storeoldfile()
            shutil.move(self.__tempfilename, self.__nexusfilename)
        except Exception as e:
            print(str(e))
            os.remove(self.__tempfilename)


class Collector(object):

    """ Collector merge images of external file-formats
    into the master NeXus file
    """

    def __init__(self, nexusfilename, compression=2,
                 skipmissing=False, storeold=False, testmode=False,
                 writer=None):
        """ The constructor creates the collector object

        :param nexusfilename: the nexus file name
        :type nexusfilename: :obj:`str`
        :param compression: compression rate
        :type compression: :obj:`int`
        :param skipmissing: if skip missing images
        :type skipmissing: :obj:`bool`
        :param storeold: if backup the input file
        :type storeold: :obj:`bool`
        :param testmode: if run in a test mode
        :type testmode: :obj:`bool`
        :param writer: the writer module
        :type writer: :obj:`str`
        """
        self.__nexusfilename = nexusfilename
        self.__compression = compression
        self.__skipmissing = skipmissing
        self.__testmode = testmode
        self.__storeold = storeold
        self.__tempfilename = None
        self.__filepattern = re.compile(".+:\\d+:\\d+")
        self.__nxsfile = None
        self.__break = False
        self.__fullfilename = None
        self.__wrmodule = None
        if writer and writer.lower() in WRITERS.keys():
            self.__wrmodule = WRITERS[writer.lower()]
        self.__siginfo = dict(
            (signal.__dict__[sname], sname)
            for sname in ('SIGINT', 'SIGHUP', 'SIGALRM', 'SIGTERM'))

        for sig in self.__siginfo.keys():
            signal.signal(sig, self._signalhandler)

    def _signalhandler(self, sig, _):
        """ signal handler

        :param sig: signal name, i.e. 'SIGINT', 'SIGHUP', 'SIGALRM', 'SIGTERM'
        :type sig: :obj:`str`
        """
        if sig in self.__siginfo.keys():
            self.__break = True
            print("terminated by %s" % self.__siginfo[sig])

    def _createtmpfile(self):
        """ creates temporary file
        """
        self.__tempfilename = self.__nexusfilename + ".__nxscollect_temp__"
        while os.path.exists(self.__tempfilename):
            self.__tempfilename += "_"
        shutil.copy2(self.__nexusfilename, self.__tempfilename)

    def _storeoldfile(self):
        """ makes back up of the input file
        """
        temp = self.__nexusfilename + ".__nxscollect_old__"
        while os.path.exists(temp):
            temp += "_"
        shutil.move(self.__nexusfilename, temp)

    @classmethod
    def _absolutefilename(cls, filename, masterfile):
        """ provides absolute image file name

        :param filename: image file name
        :type: filename: :obj:`str`
        :param masterfile: nexus file name
        :type: masterfile: :obj:`str`
        :returns: absolute image file name
        :rtype: :obj:`str`
        """
        if not os.path.isabs(filename):
            nexusfilepath = os.path.join('/', *os.path.abspath(
                masterfile).split('/')[:-1])
            filename = os.path.abspath(os.path.join(nexusfilepath, filename))
        return filename

    def _findfile(self, filename, nname=None):
        """ searches for absolute image file name

        :param filename: image file name
        :type: filename: :obj:`str`
        :param nname: hdf5 node name
        :typ nname: :obj:`str`

        :returns: absolute image file name
        :rtype: :obj:`str`
        """
        filelist = []

        if nname is not None:
            tmpfname = '%s/%s/%s' % (
                os.path.splitext(self.__nexusfilename)[0],
                nname,
                filename.split("/")[-1])
            if os.path.exists(tmpfname):
                return tmpfname
            else:
                filelist.append(tmpfname)
            tmpfname = '%s/%s/%s' % (
                os.path.splitext(self.__fullfilename)[0],
                nname,
                filename.split("/")[-1])
            if os.path.exists(tmpfname):
                return tmpfname
            else:
                filelist.append(tmpfname)
        tmpfname = self._absolutefilename(filename, self.__nexusfilename)
        if os.path.exists(tmpfname):
            return tmpfname
        else:
            filelist.append(tmpfname)
        tmpfname = self._absolutefilename(filename, self.__fullfilename)
        if os.path.exists(tmpfname):
            return tmpfname
        else:
            filelist.append(tmpfname)
        if os.path.exists(filename):
            return filename
        else:
            filelist.append(filename)
        if not self.__skipmissing:
            raise Exception(
                "Cannot open any of %s files" % sorted(set(filelist)))
        else:
            print("Cannot open any of %s files" % sorted(set(filelist)))
        return None

    def _loadrawimage(self, filename, dtype, shape=None):
        """ loads image from file

        :param filename: image file name
        :type filename: :obj:`str`
        :param dtype: field data type
        :type dtype: :obj:`str`
        :param shape: field shape
        :type shape: :obj:`list` <:obj:`int` >
        :returns: (image data, image data type, image shape)
        :rtype: (:class:`numpy.ndarray`, :obj:`str`, :obj:`list` <:obj:`int`>)
        """
        try:
            idata = None

            with open(filename, "rb") as fl:
                idata = numpy.fromfile(fl, dtype=dtype)
            if shape:
                idata = idata.reshape(shape)
            dtype = idata.dtype.__str__()
            shape = idata.shape
            if idata is not None:
                return idata, dtype, shape
            else:
                raise Exception("Cannot open a file %s" % filename)
        except Exception as e:
            print(str(e))
            if not self.__skipmissing:
                raise Exception("Cannot open a file %s" % filename)
            else:
                print("Cannot open a file %s" % filename)

            return None, None, None

    def _loadimage(self, filename):
        """ loads image from file

        :param filename: image file name
        :type filename: :obj:`str`
        :returns: (image data, image data type, image shape)
        :rtype: (:class:`numpy.ndarray`, :obj:`str`, :obj:`list` <:obj:`int`>)
        """
        try:
            dtype = None
            shape = None
            idata = None
            # filter fabio bugs
            import warnings
            warnings.filterwarnings(
                "ignore", message="numpy.dtype size changed")
            warnings.filterwarnings(
                "ignore", message="numpy.ufunc size changed")
            warnings.filterwarnings(
                "ignore",
                message="can't resolve package from __spec__ "
                "or __package__,")

            image = fabio.open(filename)
            if image:
                idata = image.data[...]
                dtype = image.data.dtype.__str__()
                shape = image.data.shape
                return idata, dtype, shape
            else:
                raise Exception("Cannot open a file %s" % filename)
            sys.stderr.write("L3\n")
        except Exception:
            if not self.__skipmissing:
                raise Exception("Cannot open a file %s" % filename)
            else:
                print("Cannot open a file %s" % filename)

            return None, None, None

    def _loadh5data(self, filename, path=None):
        """ loads image from hdf5 file

        :param filename: hdf5 image file name
        :type filename: :obj:`str`
        :param path: hdf5 field path
        :type path: :obj:`str`
        :returns: (image data, image data type, image shape)
        :rtype: (:class:`numpy.ndarray`, :obj:`str`, :obj:`list` <:obj:`int`>)
        """
        try:
            dtype = None
            shape = None
            nxsfile = filewriter.open_file(
                filename, readonly=True, writer=self.__wrmodule)
            if path:
                root = nxsfile.root()
                parent = root
                nodes = path.split("/")
                for nd in nodes:
                    if nd in parent.names():
                        parent = parent.open(nd)
                    else:
                        raise Exception(
                            "Error: path %s in % cannot be open" % (path, nd))
                image = parent
            else:
                image = nxsfile.default_field()
            if image is None:
                root = nxsfile.root()
                image = root.open("data")
            idata = image.read()
            if image is not None:
                idata = image[...]
                dtype = image.dtype
                shape = image.shape
            nxsfile.close()
            return idata, dtype, shape
        except Exception as e:
            print(str(e))
            if not self.__skipmissing:
                raise Exception("Cannot open a file %s" % filename)
            else:
                print("Cannot open a file %s" % filename)
            return None, None, None

    def _addattr(self, node, attrs):
        """ adds attributes to the parent node in nexus file

        :param node: parent hdf5 node
        :type node: parent hdf5 node
        :param attrs: dictionary with attributes
        """
        attrs = attrs or {}
        for name, (value, dtype, shape) in attrs.items():
            if not self.__testmode:
                node.attributes.create(
                    name, dtype, shape, overwrite=True)[...] = value
            print(" + add attribute: %s = %s" % (name, value))

    def _getfield(self, node, fieldname, dtype, shape, fieldattrs,
                  fieldcompression):
        """ creates a field in nexus file

        :param node: parent hdf5 node
        :type node: :class:`filewriter.FTGroup` or \
                    :class:`filewriter.FTLink`
        :param fieldname: field name
        :type fieldname: :obj:`str`
        :param dtype: field data type
        :type dtype: :obj:`str`
        :param shape: filed data shape
        :type shape: :obj:`list` <:obj:`int`>
        :param fieldattrs: dictionary with field attributes
        :type fieldattrs: :obj:`dict` <:obj:`str`, :obj:`str`>
        :param fieldcompression: field compression rate
        :type fieldcompression: :obj:`int`
        :returns: hdf5 field node
        :rtype: :class:`filewriter.FTField`
        """
        field = None
        if fieldname in node.names():
            return node.open(fieldname)
        else:
            if not self.__testmode:
                cfilter = None
                if fieldcompression:
                    opts = getcompression(fieldcompression)
                    if isinstance(opts, int):
                        cfilter = filewriter.data_filter(node)
                        cfilter.rate = opts
                    elif isinstance(opts, list) and opts:
                        cfilter = filewriter.data_filter(node)
                        cfilter.filterid = opts[0]
                        cfilter.options = tuple(opts[1:])
                if len(shape) == 2:
                    nshape = [0, shape[0], shape[1]]
                    nchunk = [1, shape[0], shape[1]]
                elif len(shape) == 3:
                    nshape = [0, shape[0], shape[1], shape[2]]
                    nchunk = [1, shape[0], shape[1], shape[2]]
                else:
                    nshape = [0, shape[0]]
                    nchunk = [1, shape[0]]
                field = node.create_field(
                    fieldname,
                    dtype,
                    shape=nshape,
                    chunk=nchunk,
                    dfilter=cfilter)
                self._addattr(field, fieldattrs)
            return field

    def _collectimages(self, files, node, fieldname=None, fieldattrs=None,
                       fieldcompression=None, datatype=None, shape=None):
        """ collects images

        :param files: a list of file strings
        :type files: :obj:`list` <:obj:`str`>
        :param node: hdf5 parent node
        :type node: :class:`filewriter.FTGroup` or \
                    :class:`filewriter.FTLink`
        :param fieldname: field name
        :type fieldname: :obj:`str`
        :param fieldattrs: dictionary with field attributes
        :type fieldattrs: :obj:`dict` <:obj:`str`, :obj:`str`>
        :param fieldcompression: field compression rate
        :type fieldcompression: :obj:`int`
        :param datatype: field data type
        :type datatype: :obj:`str`
        :param shape: field shape
        :type shape: :obj:`list` <:obj:`int` >
        """
        fieldname = fieldname or "data"
        field = None
        ind = 0
        for filestr in files:
            if self.__break:
                break
            inputfiles = filegenerator(filestr, self.__filepattern)
            for fname in inputfiles():
                if self.__break:
                    break
                npath = None
                if not datatype and \
                   ".h5://" in fname or ".nxs://" in fname:
                    fname, npath = fname.split("://", 1)
                if not self.__testmode or node is not None:
                    fname = self._findfile(fname, node.name)
                if not fname:
                    continue
                if datatype:
                    data, dtype, shape = self._loadrawimage(
                        fname, datatype, shape)
                elif fname.endswith(".h5") or fname.endswith(".nxs"):
                    try:
                        data, dtype, shape = self._loadh5data(fname, npath)
                    except Exception as e:
                        print(str(e))
                        data, dtype, shape = self._loadimage(fname)
                else:
                    data, dtype, shape = self._loadimage(fname)
                if data is not None:
                    ishape = shape
                    nrim = 1
                    if len(shape) == 3:
                        ishape = [shape[1], shape[2]]
                        nrim = shape[0]
                    if field is None:
                        if not self.__testmode or node is not None:
                            field = self._getfield(
                                node, fieldname, dtype, ishape,
                                fieldattrs, fieldcompression)
                    if field and ind == field.shape[0]:
                        if not self.__testmode:
                            if nrim == 1:
                                field.grow(0, 1)
                                field[-1, ...] = data
                            else:
                                field.grow(0, nrim)
                                field[field.shape[0]-nrim:, ...] = data
                        print(" * append %s " % (fname))
                    ind += nrim
                    if not self.__testmode:
                        self.__nxsfile.flush()

    def _inspect(self, parent, collection=False):
        """ collects recursively the all image files defined
        by hdf5 postrun fields bellow hdf5 parent node

        :param parent: hdf5 parent node
        :type parent: :class:`filewriter.FTGroup` or \
                      :class:`filewriter.FTLink`
        :param collection: if parent is of NXcollection type
        :type collection: :obj:`bool`
        """
        if hasattr(parent, "names"):
            if collection:
                if "postrun" in parent.names():
                    inputfiles = parent.open("postrun")
                    files = inputfiles.read()
                    if hasattr(files, "tolist"):
                        files = files.tolist()
                    if isinstance(files, (str, unicode)):
                        files = [files]
                    fieldname = "data"
                    fielddtype = None
                    fieldshape = None
                    fieldattrs = {}
                    fieldcompression = None
                    for at in inputfiles.attributes:
                        if at.name == "fieldname":
                            fieldname = filewriter.first(at.read())
                        elif at.name == "fieldcompression":
                            fieldcompression = filewriter.first(at.read())
                        elif at.name == "fielddtype":
                            fielddtype = filewriter.first(at.read())
                        elif at.name == "fieldshape":
                            fieldshape = json.loads(
                                filewriter.first(at.read()))
                        elif at.name.startswith("fieldattr_"):
                            atname = at.name[10:]
                            if atname:
                                fieldattrs[atname] = (
                                    at.read(), at.dtype, at.shape
                                )
                    print("populate: %s/%s with %s" % (
                        parent.parent.path, fieldname, files))
                    if fieldcompression is None:
                        fieldcompression = self.__compression
                    self._collectimages(
                        files, parent.parent, fieldname, fieldattrs,
                        fieldcompression, fielddtype, fieldshape)
            try:
                names = parent.names()
            except Exception:
                names = []
            for name in names:
                coll = False
                child = parent.open(name)
                if hasattr(child, "attributes"):
                    for at in child.attributes:
                        if at.name == "NX_class":
                            gtype = filewriter.first(at.read())
                            if gtype == 'NXcollection':
                                coll = True
                    self._inspect(child, coll)

    def _add(self, root, path, inputfiles, fieldtype=None, fieldshape=None):
        """appends specific data if path and inputfiles are given

        :param path: nexus path of the data field
        :type path: :obj:`str`
        :param inputfiles: a list of file strings
        :type inputfiles: :obj:`list` <:obj:`str`>
        :param datatype: field data type
        :type datatype: :obj:`str`
        :param shape: field shape
        :type shape: :obj:`list` <:obj:`int` >
        """
        groups = path.split("/")
        parent = root
        tgr = ""
        for gr in groups[:-1]:
            if gr:
                if ":" in gr:
                    gr, tgr = gr.split(":", 1)
                if parent is not None and gr in parent.names():
                    parent = parent.open(gr)
                else:
                    if not tgr:
                        tgr = "NX" + gr
                    if not self.__testmode:
                        parent = parent.create_group(gr, tgr)
                    else:
                        parent = None
                    # raise Exception(
                    #     "Error: path %s in % cannot be open" % (path, gr))

        fieldname = groups[-1]
        if parent:
            print("populate: %s/%s with %s" %
                  (parent.path, fieldname, inputfiles))
        else:
            print("populate: %s/%s with %s" %
                  (path, fieldname, inputfiles))
        fieldcompression = self.__compression
        fieldattrs = {}
        self._collectimages(
            inputfiles, parent, fieldname, fieldattrs,
            fieldcompression, fieldtype, fieldshape)

    def collect(self, path=None, inputfiles=None, datatype=None, shape=None):
        """ creates a temporary file,
        collects the all image files defined by hdf5
        postrun fields of NXcollection groups and renames the temporary file
        to the origin one if the action was successful
        or appends specific data if path and inputfiles are given

        :param path: nexus path of the data field
        :type path: :obj:`str`
        :param inputfiles: a list of file strings
        :type inputfiles: :obj:`list` <:obj:`str`>
        :param datatype: field data type
        :type datatype: :obj:`str`
        :param shape: field shape
        :type shape: :obj:`list` <:obj:`int` >
        """
        self._createtmpfile()
        try:
            self.__nxsfile = filewriter.open_file(
                self.__tempfilename, readonly=self.__testmode,
                writer=self.__wrmodule)
            root = self.__nxsfile.root()
            try:
                self.__fullfilename = filewriter.first(
                    root.attributes['file_name'].read())
                # print self.__fullfilename
            except Exception:
                pass
            if path and inputfiles:
                self._add(root, path, inputfiles, datatype, shape)
            else:
                self._inspect(root)
            self.__nxsfile.close()
            if self.__storeold:
                self._storeoldfile()
            shutil.move(self.__tempfilename, self.__nexusfilename)
        except Exception as e:
            print(str(e))
            os.remove(self.__tempfilename)


class VDS(Runner):

    """ Execute runner
    """

    #: (:obj:`str`) command description
    description = "create a virual dataset in the master file"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n\n" \
        + "       nxscollect vds " \
        + "scan_234.nxs://entry/instrument/eiger/data " \
        + " --shape '1000,2048,1024' --dtype uint32 " \
        + " --target-fields 'eiger_%05d.nxs://entry/data/data:1:10'" \
        + " --shapes '100,,:100,,:100,,:100,,:100,,:100,,:100,," \
        + ":100,,:100,,:100,,'  " \
        + " --offsets '0,,:100,,:200,,:300,,:400,,:500,," \
        + ":600,,:700,,:800,,:900,,' \n\n" \
        + "\n" \
        + "           - creates VDS (shape [1000,2048,1024]) of" \
        " ten nexus files (shape [100,2048,1024])" \
        + " merged in their first dimension\n" \
        + "\n\n\n" \
        + "       nxscollect vds " \
        + "scan_234.nxs://entry/instrument/lambda/data " \
        + " --shape '100,300,762' --dtype uint32 " \
        + " --target-fields 'lambda_%05d.nxs://entry/data/data:0:2'" \
        + " --shapes ',,250:,,250:,,250'  " \
        + " --offsets ',,:,,256:,,512'  --counts 'U,,:U,,:U,,' -f 1 \n\n" \
        + "\n" \
        + "           - creates VDS (shape [100,300,762]) of" \
        " three nexus files (shape [100,300,250])" \
        + " merged in their third dimension,\n" \
        + "               separated with a 6 pixel gap of 1 values" \
        + " and unlimited first dimension\n" \
        + "\n\n\n" \
        + "       nxscollect vds " \
        + "scan_234.nxs://entry/instrument/percival/data " \
        + " --shape '4000,1600,2000' --dtype int16 " \
        + " --target-fields 'percival_%05d.nxs://entry/data/data:1:4'" \
        + " --shapes '1000,,:1000,,:1000,,:1000,,'  " \
        + " --offsets '0,,:1,,:2,,:3,,'  --counts 'U,,:U,,:U,,:U,,'" \
        + " --strides '4,,:4,,:4,,:4,,' \n\n" \
        + "\n" \
        + "           - creates VDS (shape [1000,1600,2000]) of" \
        " three nexus files (shape [1000,1600,2000])\n" \
        + "                merged in their the first dimension " \
        + "with interlaying frames\n" \
        + "                and unlimited first dimension\n" \
        + "\n\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-t", "--dtype", dest="dtype",
            action="store", type=str, default=None,
            help="datatype of the VDS field, e.g. 'uint8'")
        parser.add_argument(
            "-s", "--shape", dest="shape",
            action="store", type=str, default=None,
            help="shape of the VDS field, e.g. '[U,4096,2048]' or U,4096,2048"
            " where U means span along the field'")
        parser.add_argument(
            "-f", "--fill-value", dest="fillvalue",
            action="store", type=str, default=None,
            help="fill value for the gaps, default is 0")
        parser.add_argument(
            "-e", "--target-fields", dest="targetfields",
            action="store", type=str, default=None,
            help="external fields with their NeXus file paths "
            "defined with a pattern or separated by ',' e.g."
            "'scan_123/lambda_%%05d.nxs://entry/data/data:0:3'")
        parser.add_argument(
            "--separator", dest="separator",
            action="store", type=str, default=",",
            help="input data files separator (default: ',')")
        parser.add_argument(
            "-p", "--shapes", dest="shapes",
            action="store", type=str, default=None,
            help="shapes in the VDS layout hyperslab "
            "for the corresponding target fields  "
            "with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            "',,;,300,;,600,0' where an empty coordinate means 0")
        parser.add_argument(
            "-o", "--offsets", dest="offsets",
            action="store", type=str, default=None,
            help="offsets in the VDS layout hyperslab "
            "for the corresponding target fields  "
            "with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            "',,;,300,;,600,0' where an empty coordinate means 0")
        parser.add_argument(
            "-b", "--blocks", dest="blocks",
            action="store", type=str, default=None,
            help="block sizes in the VDS layout hyperslab "
            "for the corresponding target fields  "
            " with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            " ',256,512;,256,512;,256,512' "
            "where an empty coordinate means 1")
        parser.add_argument(
            "-c", "--counts", dest="counts",
            action="store", type=str, default=None,
            help="count numbers in the VDS layout hyperslab"
            "for the corresponding target fields  "
            " with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            " ',1,1;,1,1;,1,1' "
            "where an empty coordinate means span along the layout")
        parser.add_argument(
            "-d", "--strides", dest="strides",
            action="store", type=str, default=None,
            help="stride sizes in the VDS layout hyperslab"
            "for the corresponding target fields  "
            " with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            " ',,;,,;,,' "
            "where an empty coordinate means 1")
        parser.add_argument(
            "-l", "--slices", dest="slices",
            action="store", type=str, default=None,
            help="mapping slices in the VDS layout"
            "for the corresponding target fields  "
            " with coordinates sepatated by ',' "
            "and different fields separated by ';'  or spaces e.g."
            " ':,0:50,: :,50:100,:' "
            "where U means span along the layout ")
        parser.add_argument(
            "-P", "--target-shapes", dest="targetshapes",
            action="store", type=str, default=None,
            help="field shapes "
            "with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            "',,;,300,;,600,0'")
        parser.add_argument(
            "-O", "--target-offsets", dest="targetoffsets",
            action="store", type=str, default=None,
            help="offsets in the view hyperslab of target fields"
            "with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            "',,;,300,;,600,0' where an empty coordinate means 0")
        parser.add_argument(
            "-B", "--target-blocks", dest="targetblocks",
            action="store", type=str, default=None,
            help="block sizes in the view hyperslab of target fields"
            " with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            " ',256,512;,256,512;,256,512' "
            "where an empty coordinate means 1")
        parser.add_argument(
            "-C", "--target-counts", dest="targetcounts",
            action="store", type=str, default=None,
            help="count numbers in the view hyperslab of target fields"
            " with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            " ',1,1;,1,1;,1,1' "
            "where an empty coordinate means span along the layout")
        parser.add_argument(
            "-D", "--target-strides", dest="targetstrides",
            action="store", type=str, default=None,
            help="stride sizes numbers in the view hyperslab "
            "of target fields"
            " with coordinates sepatated by ',' "
            "and different fields separated by ';', ':' or spaces e.g."
            " ',,;,,;,,' "
            "where an empty coordinate means 1")
        parser.add_argument(
            "-L", "--target-slices", dest="targetslices",
            action="store", type=str, default=None,
            help="view slices of target fields"
            " with coordinates sepatated by ',' "
            "and different fields separated by ';'  or spaces e.g."
            " ':,0:50,: :,0:50,:' "
            "where U means span along the layout ")
        parser.add_argument(
            "-r", "--replace-nexus-file", action="store_true",
            default=False, dest="replaceold",
            help="if it is set the old file is not copied into "
            "a file with .__nxscollect__old__* extension")
        parser.add_argument(
            "--test", action="store_true",
            default=False, dest="testmode",
            help="execute in the test mode")
        parser.add_argument(
            "--h5cpp", action="store_true",
            default=False, dest="h5cpp",
            help="use h5cpp module as a nexus reader")
        parser.add_argument(
            "--h5py", action="store_true",
            default=False, dest="h5py",
            help="use h5py module as a nexus reader/writer")

    def postauto(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            'args', metavar='nexus_file_path_field',
            type=str, nargs='?',
            help='nexus files with the nexus directory and a field name '
            'to create the VDS field')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        parser = self._parser
        nexusfilepath = options.args

        if not nexusfilepath or not nexusfilepath[0]:
            parser.print_help()
            print("")
            sys.exit(0)

        if options.targetfields is None:
            sys.stderr.write("nxscollect: target fields are missing\n")
            parser.print_help()
            print("")
            sys.exit(0)

        if options.shape is None:
            sys.stderr.write("nxscollect: shape is missing\n")
            parser.print_help()
            print("")
            sys.exit(0)

        if options.shapes is None:
            sys.stderr.write("nxscollect: shapes is missing\n")
            parser.print_help()
            print("")
            sys.exit(0)

        if options.h5cpp:
            writer = "h5cpp"
        elif options.h5py:
            writer = "h5py"
        elif "h5cpp" in WRITERS.keys():
            writer = "h5cpp"
        else:
            writer = "h5py"

        if (options.h5py and options.h5cpp) or \
           writer not in WRITERS.keys():
            sys.stderr.write("nxscollect: Writer '%s' cannot be opened\n"
                             % writer)
            sys.stderr.flush()
            parser.print_help()
            sys.exit(255)

        # configuration server
        vds = VirtualDataset(
            nexusfilepath, options, writer=writer)
        vds.create()


class Link(Runner):

    """ Execute runner
    """

    #: (:obj:`str`) command description
    description = "create an external or internal link in the master file"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxscollect link " \
        + "scan_234.nxs://entry/instrument/lambda " \
        + "--name data --target scan_234/lambda.nxs://entry/data/data \n\n" \
        + "scan_234.nxs://entry/instrument/eiger:NXdetector " \
        + "  --target scan_234/eiger.nxs://entry/data/data \n\n" \
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-n", "--name", dest="name",
            action="store", type=str, default=None,
            help="link name")
        parser.add_argument(
            "-t", "--target", dest="target",
            action="store", type=str, default=None,
            help="link target with the file name if external")
        parser.add_argument(
            "-r", "--replace-nexus-file", action="store_true",
            default=False, dest="replaceold",
            help="if it is set the old file is not copied into "
            "a file with .__nxscollect__old__* extension")
        parser.add_argument(
            "--test", action="store_true",
            default=False, dest="testmode",
            help="execute in the test mode")
        parser.add_argument(
            "--h5cpp", action="store_true",
            default=False, dest="h5cpp",
            help="use h5cpp module as a nexus reader")
        parser.add_argument(
            "--h5py", action="store_true",
            default=False, dest="h5py",
            help="use h5py module as a nexus reader/writer")

    def postauto(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            'args', metavar='nexus_file_path',
            type=str, nargs='?',
            help='nexus files with the nexus directory to place the link')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        parser = self._parser
        nexusfilepath = options.args

        if not nexusfilepath or not nexusfilepath[0]:
            parser.print_help()
            print("")
            sys.exit(0)

        if options.h5cpp:
            writer = "h5cpp"
        elif options.h5py:
            writer = "h5py"
        elif "h5cpp" in WRITERS.keys():
            writer = "h5cpp"
        else:
            writer = "h5py"

        if (options.h5py and options.h5cpp) or \
           writer not in WRITERS.keys():
            sys.stderr.write("nxscollect: Writer '%s' cannot be opened\n"
                             % writer)
            sys.stderr.flush()
            parser.print_help()
            sys.exit(255)

        # configuration server
        linker = Linker(
            nexusfilepath, options.target, options.name,
            not options.replaceold, options.testmode, writer=writer)
        linker.link()


class Execute(Runner):

    """ Execute runner
    """

    #: (:obj:`str`) command description
    description = "append images to the master file"
    #: (:obj:`str`) command epilog
    epilog = "" \
        + " examples:\n" \
        + "       nxscollect append -c1 /tmp/gpfs/raw/scan_234.nxs \n\n" \
        + "       nxscollect append -c32008:0,2 /ramdisk/scan_123.nxs \n\n" \
        + "       nxscollect append --test /tmp/gpfs/raw/scan_234.nxs \n\n" \
        + "       nxscollect append scan_234.nxs " \
        + "--path /scan/instrument/pilatus/data  " \
        + "--input-files 'scan_%05d.tif:0:100' "\
        + "\n"

    def create(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument(
            "-c", "--compression", dest="compression",
            action="store", type=str, default="2",
            help="deflate compression rate from 0 to 9 (default: 2)"
            " or <filterid>:opt1,opt2,..."
            " e.g.  -c 32008:0,2  for bitshuffle with lz4")
        parser.add_argument(
            "-p", "--path", dest="path",
            action="store", type=str, default=None,
            help="nexus path for the output field, e.g."
            " /scan/instrument/pilatus/data")
        parser.add_argument(
            "-i", "--input-files", dest="inputfiles",
            action="store", type=str, default=None,
            help="input data files defined with a pattern "
            "or separated by ',' e.g."
            "'scan_%%05d.tif:0:100'")
        parser.add_argument(
            "--separator", dest="separator",
            action="store", type=str, default=",",
            help="input data files separator (default: ',')")
        parser.add_argument(
            "--dtype", dest="datatype",
            action="store", type=str, default=None,
            help="datatype of input data - only for raw data,"
            " e.g. 'uint8'")
        parser.add_argument(
            "--shape", dest="shape",
            action="store", type=str, default=None,
            help="shape of input data - only for raw data,"
            " e.g. '[4096,2048]'")
        parser.add_argument(
            "-s", "--skip-missing", action="store_true",
            default=False, dest="skipmissing",
            help="skip missing files")
        parser.add_argument(
            "-r", "--replace-nexus-file", action="store_true",
            default=False, dest="replaceold",
            help="if it is set the old file is not copied into "
            "a file with .__nxscollect__old__* extension")
        parser.add_argument(
            "--test", action="store_true",
            default=False, dest="testmode",
            help="execute in the test mode")
        parser.add_argument(
            "--h5cpp", action="store_true",
            default=False, dest="h5cpp",
            help="use h5cpp module as a nexus reader")
        parser.add_argument(
            "--h5py", action="store_true",
            default=False, dest="h5py",
            help="use h5py module as a nexus reader/writer")

    def postauto(self):
        """ creates parser
        """
        parser = self._parser
        parser.add_argument('args', metavar='nexus_file',
                            type=str, nargs='*',
                            help='nexus files to be collected')

    def run(self, options):
        """ the main program function

        :param options: parser options
        :type options: :class:`argparse.Namespace`
        """
        parser = self._parser
        nexusfiles = options.args

        try:
            getcompression(options.compression)
        except Exception as e:
            print(str(e))
            parser.print_help()
            print("")
            sys.exit(0)

        if not nexusfiles or not nexusfiles[0]:
            parser.print_help()
            print("")
            sys.exit(0)

        if options.h5cpp:
            writer = "h5cpp"
        elif options.h5py:
            writer = "h5py"
        elif "h5cpp" in WRITERS.keys():
            writer = "h5cpp"
        else:
            writer = "h5py"
        if (options.h5py and options.h5cpp) or \
           writer not in WRITERS.keys():
            sys.stderr.write("nxscollect: Writer '%s' cannot be opened\n"
                             % writer)
            sys.stderr.flush()
            parser.print_help()
            sys.exit(255)
        if (options.path and not options.inputfiles):
            sys.stderr.write(
                "nxscollect: --input-files argument is missing")
            parser.print_help()
            sys.exit(255)
        if (not options.path and options.inputfiles):
            sys.stderr.write(
                "nxscollect: --path argument is missing")
            parser.print_help()
            sys.exit(255)
        inputfiles = None
        if options.inputfiles:
            if options.separator:
                inputfiles = options.inputfiles.split(options.separator)
            else:
                inputfiles = [options.inputfiles]

        shape = None
        if options.shape:
            try:
                shape = json.loads(options.shape)
            except Exception:
                sys.stderr.write(
                    "nxscollect: shape is not readable")
                parser.print_help()
                sys.exit(255)

        # configuration server
        for nxsfile in nexusfiles:
            collector = Collector(
                nxsfile, options.compression, options.skipmissing,
                not options.replaceold, options.testmode, writer=writer)
            collector.collect(options.path, inputfiles,
                              options.datatype, shape)


def _supportoldcommands():
    """ replace the old command names to the new ones
    """

    oldnew = {
        '-x': 'append',
        '--execute': 'append',
        '--replace_nexus_file': '--replace-nexus-file',
        '--input_files': '--input-files',
        '--skip_missing': '--skip-missing',
        'execute': 'append',
    }

    if sys.argv and len(sys.argv) > 1:
        for i, arg in enumerate(sys.argv):
            if i > 0 and arg in oldnew.keys():
                sys.stderr.write(
                    "Warning: `%s` is deprecated, "
                    "please use `%s` instead\n\n"
                    % (arg, oldnew[arg]))
                sys.argv[i] = oldnew[arg]


def main():
    """ the main program function
    """
    description = "  Command-line tool to merge images of external " \
                  + "file-formats into the master NeXus file"

    epilog = 'For more help:\n  nxscollect <sub-command> -h'
    _supportoldcommands()

    parser = NXSArgParser(
        description=description, epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.cmdrunners = [
        ('append', Execute),
        ('link', Link),
        ('vds', VDS)
    ]
    runners = parser.createSubParsers()

    try:
        options = parser.parse_args()
    except ErrorException as e:
        sys.stderr.write("Error: %s\n" % str(e))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    if options.subparser is None:
        sys.stderr.write(
            "Error: %s\n" % str("too few arguments"))
        sys.stderr.flush()
        parser.print_help()
        print("")
        sys.exit(255)

    if not WRITERS:
        sys.stderr.write(
            "nxsfileinfo: Neither pnineuxs.h5cpp nor h5py installed\n")
        sys.stderr.flush()
        parser.print_help()
        sys.exit(255)

    runners[options.subparser].run(options)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

""" Command-line tool for creating to the nexdatas configuration server """

import copy
import os.path
import json
import sys

import lxml.etree as etree
import xml.etree.ElementTree as et
from lxml.etree import XMLParser

from nxstools.nxsdevicetools import (
    storeDataSource, getDataSourceComponents, storeComponent,
    moduleAttributes, motorModules,
    generateDeviceNames, getServerTangoHost,
    openServer, findClassName,
    xmlPackageHandler)
from nxstools.nxsxml import (XMLFile, NDSource, NGroup, NField, NLink,
                             NDimensions)

#: (:obj:`bool`) True if PyTango available
PYTANGO = False
try:
    import PyTango
    PYTANGO = True
except Exception:
    pass

if sys.version_info > (3,):
    basestring = str
    unicode = str


class CPExistsException(Exception):

    """ Component already exists exception
    """


class DSExistsException(Exception):

    """ DataSource already exists exception
    """


def _tostr(text):
    """ converts text  to str type

    :param text: text
    :type text: :obj:`bytes` or :obj:`unicode`
    :returns: text in str type
    :rtype: :obj:`str`
    """
    if hasattr(text, "tostring"):
        text = text.tostring()
    if isinstance(text, str):
        return text
    elif sys.version_info > (3,):
        return str(text, encoding="utf8")
    else:
        return str(text)


def _toxml(node):
    """ provides xml content of the whole node

    :param node: DOM node
    :type node: :class:`xml.dom.Node`
    :returns: xml content string
    :rtype: :obj:`str`
    """
    xml = _tostr(et.tostring(node, encoding='utf8', method='xml'))
    if xml.startswith("<?xml version='1.0' encoding='utf8'?>"):
        xml = str(xml[38:])
    return xml


def _simpletoxml(node):
    """ provides xml content of the whole node

    :param node: DOM node
    :type node: :class:`xml.dom.Node`
    :returns: xml content string
    :rtype: :obj:`str`
    """
    return _tostr(et.tostring(node, encoding='utf8', method='xml'))


class Device(object):

    """ device from online.xml
    """
    __slots__ = [
        'name', 'dtype', 'module', 'tdevice', 'sdevice',
        'hostname', 'sardananame', 'sardanahostname',
        'host', 'port', 'shost', 'sport', 'thost', 'tport',
        'group', 'attribute']

    def __init__(self):
        #: (:obj:`str`) device name
        self.name = None
        #: (:obj:`str`) device type
        self.dtype = None
        #: (:obj:`str`) device module
        self.module = None
        #: (:obj:`str`) tango device name
        self.tdevice = None
        #: (:obj:`str`) sardana device name
        self.sdevice = None
        #: (:obj:`str`) host name with port
        self.hostname = None
        #: (:obj:`str`) sardana name with port
        self.sardananame = None
        #: (:obj:`str`) sardana host name
        self.sardanahostname = None
        #: (:obj:`str`) host without port
        self.host = None
        #: (:obj:`str`) port
        self.port = None
        #: (:obj:`str`) tango host without port
        self.thost = None
        #: (:obj:`str`) tango port
        self.tport = None
        #: (:obj:`str`) sardana host without port
        self.shost = None
        #: (:obj:`str`) sardana tango port
        self.sport = None
        #: (:obj:`str`) datasource tango group
        self.group = None
        #: (:obj:`str`) attribute name
        self.attribute = None

    def compare(self, dv):
        dct = {}
        tocompare = [
            "name", "dtype", "module", "tdevice", "hostname",
            "sardananame", "sardanahostname"]
        for at in tocompare:
            v1 = getattr(self, at)
            v2 = getattr(dv, at)
            if v1 != v2:
                dct[at] = (str(v1) if v1 else v1, str(v2) if v2 else v2)
        return dct

    def tolower(self):
        """ converts `name`, `module`, `tdevice`, `hostname` into lower case
        """
        if self.name:
            self.name = self.name.lower()
        if self.module:
            self.module = self.module.lower()
        if self.tdevice:
            self.tdevice = self.tdevice.lower()
        if self.hostname:
            self.hostname = self.hostname.lower()

    def splitHostPort(self):
        """ spilts host name from port
        """
        if self.hostname:
            self.host = self.hostname.split(":")[0]
            self.port = self.hostname.split(":")[1] \
                if len(self.hostname.split(":")) > 1 else None
            self.thost = self.host
            self.tport = self.port
        else:
            self.host = None
            self.port = None
            raise Exception("hostname not defined")

    def findDevice(self, tangohost):
        """ sets sardana device name and sardana host/port of online.xml device

        :param tangohost: tango host
        :type tangohost: :obj:`str`
        """
        mhost = self.sardanahostname or tangohost
        self.sdevice = None
        self.shost = None
        self.sport = None
        if PYTANGO:
            try:
                dp = PyTango.DeviceProxy(str("%s/%s" % (mhost, self.name)))
                mdevice = str(dp.name())

                #  self.hostname = mhost
                self.shost = mhost.split(":")[0]
                if len(mhost.split(":")) > 1:
                    self.sport = mhost.split(":")[1]

                self.sdevice = mdevice
            except Exception:
                pass

    def findAttribute(self, tangohost, clientlike=False):
        """ sets attribute and datasource group of online.xml device

        :param tangohost: tango host
        :type tangohost: :obj:`str`
        :param clientlike: tango motors to be client like
        :type clientlike: :obj:`bool`
        """
        mhost = self.sardanahostname or tangohost
        self.group = None
        self.attribute = None
        spdevice = self.tdevice.split("/")
        if mhost and len(spdevice) > 3:
            self.attribute = spdevice[3]
            self.tdevice = "/".join(spdevice[0:3])
            self.hostname = mhost
        if self.module in motorModules or self.dtype == 'stepping_motor':
            if self.attribute is None:
                self.attribute = 'Position'
            if clientlike:
                self.group = '__CLIENT__'
        elif PYTANGO and self.module in moduleAttributes:
            try:
                try:
                    dp = PyTango.DeviceProxy(
                        str("%s/%s" % (mhost, self.sardananame)))
                except Exception:
                    dp = PyTango.DeviceProxy(str("%s/%s" % (mhost, self.name)))
                mdevice = str(dp.name())

                sarattr = moduleAttributes[self.module][0]
                if not sarattr or \
                   sarattr not in dp.get_attribute_list():
                    raise Exception("Missing attribute: Value")
                self.hostname = mhost
                self.host = mhost.split(":")[0]
                self.shost = self.host
                if len(mhost.split(":")) > 1:
                    self.port = mhost.split(":")[1]
                    self.sport = self.port
                if mdevice:
                    self.sdevice = mdevice
                self.attribute = sarattr
                self.group = '__CLIENT__'
            except Exception:
                if moduleAttributes[self.module][1]:
                    if self.attribute is None:
                        self.attribute = moduleAttributes[self.module][1]
                    self.group = '__CLIENT__'

    def setSardanaName(self, tolower):
        """ sets sardana name

        :param tolower: If True name in lowercase
        :type tolower: :obj:`bool`
        """
        self.name = self.sardananame or self.name
        if tolower:
            self.name = self.name.lower()


class Creator(object):

    """ configuration server adapter
    """

    def __init__(self, options, args, printouts=True):
        """ constructor

        :param options:  command options
        :type options: :class:`optparse.Values`
        :param args: command arguments
        :type args: :obj:`list` < :obj:`str` >
        :param printouts: if printout is enable
        :type printouts: :obj:`bool`
        """
        #: (:class:`optparse.Values`) creator options
        self.options = options
        #: (:obj:`list` < :obj:`str` >) creator arguments
        self.args = args
        #: (:obj:`bool`) if printout is enable
        self._printouts = printouts

    @classmethod
    def _areComponentsAvailable(cls, names, server, lower=False):
        """ checks if the components are available

        :param names: component names
        :type names: :obj:`list` < :obj:`str` >
        :param server: server name
        :type server: :obj:`str`
        :param lower: checks lower case name
        :type lower: :obj:`bool`
        :returns: a subset of available components
        :rtype:  :obj:`list` < :obj:`str` >
        """
        try:
            proxy = openServer(server)
            proxy.Open()
            acps = proxy.availableComponents()
        except Exception:
            raise Exception("Cannot connect to %s" % server)
        cps = []
        for name in names:
            if name in acps or (lower and name.lower() in acps):
                cps.append(name)
        return cps

    @classmethod
    def _componentFilesExist(cls, names, fileprefix, directory):
        """ checks if the components files exist

        :param names: component names
        :type names: :obj:`list` < :obj:`str` >
        :param fileprefix: file name prefix
        :type fileprefix: :obj:`str`
        :param directory: file directory
        :type directory: :obj:`str`
        :param lower: checks lower case name
        :type lower: :obj:`bool`
        :returns: a subset of available components
        :rtype:  :obj:`list` < :obj:`str` >
        """
        cps = []
        for name in names:
            fname = "%s/%s%s.xml" % (directory, fileprefix, name)
            if os.path.exists(fname):
                cps.append(fname)
        return cps

    @classmethod
    def _areDataSourcesAvailable(cls, names, server, lower=False):
        """ checks if the datasources are available

        :param names: datasource names
        :type names: :obj:`list` < :obj:`str` >
        :param server: server name
        :type server: :obj:`str`
        :param lower: checks lower case name
        :type lower: :obj:`bool`
        :returns: a subset of available datasources
        :rtype:  :obj:`list` < :obj:`str` >
        """
        try:
            proxy = openServer(server)
            proxy.Open()
            adss = proxy.availableDataSources()
        except Exception:
            raise Exception("Cannot connect to %s" % server)
        dss = []
        for name in names:
            if name in adss or (lower and name.lower() in adss):
                dss.append(name)
        return dss

    @classmethod
    def _dataSourceFilesExist(cls, names, fileprefix, directory):
        """ checks if the datasources files exist

        :param names: datasource names
        :type names: :obj:`list` < :obj:`str` >
        :param fileprefix: file name prefix
        :type fileprefix: :obj:`str`
        :param directory: file directory
        :type directory: :obj:`str`
        :param lower: checks lower case name
        :type lower: :obj:`bool`
        :returns: a subset of available datasources
        :rtype:  :obj:`list` < :obj:`str` >
        """
        dss = []
        for name in names:
            fname = "%s/%s%s.ds.xml" % (directory, fileprefix, name)
            if os.path.exists(fname):
                dss.append(fname)
        return dss

    @classmethod
    def _createTangoDataSource(
            cls, name, directory, fileprefix, server, device,
            elementname, host, port="10000", group=None, elementtype=None):
        """ creates TANGO datasource file

        :param name: device name
        :type name: :obj:`str`
        :param directory: output file directory
        :type directory: :obj:`str`
        :param fileprefix: file name prefix
        :type fileprefix: :obj:`str`
        :param server: server name
        :type server: :obj:`str`
        :param device: device name
        :type device: :obj:`str`
        :param elementname: element name, e.g. attribute name
        :type elementname: :obj:`str`
        :param host: tango host name
        :type host: :obj:`str`
        :param port: tango port
        :type port: :obj:`str`
        :parma group: datasource tango group
        :type group: :obj:`str`
        :parma elementtype: element type, i.e. attribute, property or command
        :type elementtype: :obj:`str`
        :returns: xml string
        :rtype: :obj:`str`
        """
        df = XMLFile("%s/%s%s.ds.xml" % (directory, fileprefix, name))
        sr = NDSource(df)
        sr.initTango(name, device, elementtype or "attribute",
                     elementname, host, port, group=group)
        xml = df.prettyPrint()
        if server:
            storeDataSource(name, xml, server)
        elif directory is not None and fileprefix is not None:
            df.dump()
        return xml

    @classmethod
    def _createClientDataSource(
            cls, name, directory, fileprefix, server, dsname=None):
        """ creates CLIENT datasource file

        :param name: device name
        :type name: :obj:`str`
        :param directory: output file directory
        :type directory: :obj:`str`
        :param fileprefix: file name prefix
        :type fileprefix: :obj:`str`
        :param server: server name
        :type server: :obj:`str`
        :param dsname: datasource name
        :type dsname: :obj:`str`
        :returns: xml string
        :rtype: :obj:`str`
        """
        dname = name if not dsname else dsname
        df = XMLFile("%s/%s%s.ds.xml" % (directory, fileprefix, dname))
        print("%s/%s%s.ds.xml" % (directory, fileprefix, dname))
        sr = NDSource(df)
        sr.initClient(dname, name)
        xml = df.prettyPrint()
        if server:
            storeDataSource(dname, xml, server)
        elif directory is not None and fileprefix is not None:
            df.dump()
        return xml

    @classmethod
    def __patheval(cls, nexuspath):
        """ splits nexus path into list

        :param nexuspath: nexus path
        :type nexuspath: :obj:`str`
        :returns: nexus path in lists of (name, NXtype)
        :rtype: :obj:`list` < (:obj:`str`, :obj:`str`) >
        """
        pathlist = []
        spath = nexuspath.split("/")
        if spath:
            for sp in spath[:-1]:
                nlist = sp.split(":")
                if len(nlist) == 2:
                    if len(nlist[0]) == 0 and \
                       len(nlist[1]) > 2 and nlist[1].startswith("NX"):
                        pathlist.append((nlist[1][2:], nlist[1]))
                    else:
                        pathlist.append((nlist[0], nlist[1]))
                elif len(nlist) == 1 and nlist[0]:
                    if len(nlist[0]) > 2 and nlist[0].startswith("NX"):
                        pathlist.append((nlist[0][2:], nlist[0]))
                    else:
                        pathlist.append((nlist[0], "NX" + nlist[0]))

            pathlist.append((spath[-1], None))
        return pathlist

    @classmethod
    def __createTree(cls, df, nexuspath, name, nexusType,
                     strategy, units, link, chunk, canfail=None):
        """ create nexus node tree

        :param df: definition parent node
        :type df: :class:'nxstools.nxsxml.XMLFile'
        :param nexuspath: nexus path
        :type nexuspath: :obj:`str`
        :param name: name
        :type name: :obj:`str`
        :param nexusType: nexus type
        :type nexusType: :obj:`str`
        :param strategy: strategy mode
        :type startegy: :obj:`str`
        :param units: field units
        :type units: :obj:`str`
        :param links: if create link
        :type links: :obj:`bool`
        :param chunk: chunk size, e.g. `SCALAR`, `SPECTRUM` or `IMAGE`
        :type chunk: :obj:`str`
        :param canfail: can fail strategy flag
        :type canfail: :obj:`bool`
        """

        pathlist = cls.__patheval(nexuspath)
        entry = None
        parent = df
        for path in pathlist[:-1]:
            child = NGroup(parent, path[0], path[1])
            if parent == df:
                entry = child
            parent = child
        if pathlist:
            fname = pathlist[-1][0] or name
            field = NField(parent, fname, nexusType)
            field.setStrategy(strategy, canfail=canfail)
            if units.strip():
                field.setUnits(units.strip())
            field.setText("$datasources.%s" % name)
            if chunk != 'SCALAR':
                if chunk == 'SPECTRUM':
                    NDimensions(field, "1")
                elif chunk == 'IMAGE':
                    NDimensions(field, "2")
            if link and entry:
                npath = (nexuspath + name) \
                    if nexuspath[-1] == '/' else nexuspath
                data = NGroup(entry, "data", "NXdata")
                if link > 1:
                    NLink(data, name, npath)
                else:
                    NLink(data, fname, npath)

    @classmethod
    def _createComponent(cls, name, directory, fileprefix, nexuspath,
                         strategy, nexusType, units, links, server, chunk,
                         dsname, canfail=None):
        """ creates component file

        :param name: component name
        :type name: :obj:`str`
        :param directory: output file directory
        :type directory: :obj:`str`
        :param fileprefix: file name prefix
        :type fileprefix: :obj:`str`
        :param nexuspath: nexus path
        :type nexuspath: :obj:`str`
        :param strategy: field strategy
        :type startegy: :obj:`str`
        :param nexusType: nexus Type of the field
        :type nexusType: :obj:`str`
        :param units: field units
        :type units: :obj:`str`
        :param link: nxdata link
        :type links: :obj:`bool`
        :param server: configuration server
        :type server: :obj:`str`
        :returns: component xml
        :rtype: :obj:`str`
        :param dsname: datasource name
        :type dsname: :obj:`str`
        :param canfail: can fail strategy flag
        :type canfail: :obj:`bool`
        """
        defpath = "/$var.entryname#'scan'$var.serialno:NXentry/instrument" \
                  + "/collection/%s" % (dsname or name)
        df = XMLFile("%s/%s%s.xml" % (directory, fileprefix, name))
        cls.__createTree(df, nexuspath or defpath, dsname or name, nexusType,
                         strategy, units, links, chunk, canfail)

        xml = df.prettyPrint()
        if server:
            storeComponent(name, xml, server)
        elif directory is not None and fileprefix is not None:
            df.dump()
        return xml

    @classmethod
    def _getText(cls, node):
        """ provides xml content of the node

        :param node: DOM node
        :type node: :class:`lxml.etree.Element`
        :returns: xml content string
        :rtype: :obj:`str`
        """
        if node is None:
            return
        xml = _toxml(node)
        start = xml.find('>')
        end = xml.rfind('<')
        if start == -1 or end < start:
            return ""
        return xml[start + 1:end].replace("&lt;", "<").replace("&gt;", "<"). \
            replace("&quot;", "\"").replace("&amp;", "&")

    @classmethod
    def _getChildText(cls, parent, childname):
        """ provides text of child named by childname

        :param parent: parent node
        :type parent: :class:`lxml.etree.Element`
        :param childname: child name
        :type childname: :opj:`str`
        :returns: text string
        :rtype: :obj:`str`
        """
        children = parent.findall(childname)
        return cls._getText(children[0]) if len(children) else None

    def _getModuleName(self, device):
        """ provides module name

        :param device: device name
        :type device: :obj:`str`
        :returns: module name
        :rtype: :obj:`str`
        """
        if device.module.lower() in \
           self.xmlpackage.moduleMultiAttributes.keys():
            return device.module.lower()
        elif len(device.tdevice.split('/')) == 3:
            try:
                classname = findClassName(device.hostname, device.tdevice)
                if classname.lower() \
                   in self.xmlpackage.moduleMultiAttributes.keys():
                    return classname.lower()
            except Exception:
                pass
            if device.module.lower() == 'module_tango' \
               and len(device.tdevice.split('/')) == 3 \
               and device.tdevice.split('/')[1] \
               in self.xmlpackage.moduleMultiAttributes.keys():
                return device.tdevice.split('/')[1].lower()


class WrongParameterError(Exception):

    """ wrong parameter exception
    """
    pass


class ComponentCreator(Creator):

    """ component creator
    """

    def create(self):
        """ creates a component xml and stores it in DB or filesytem
        """
        aargs = []
        if self.options.device.strip():
            try:
                first = int(self.options.first)
            except Exception:
                raise WrongParameterError(
                    "ComponentCreator Invalid --first parameter\n")

            try:
                last = int(self.options.last)
            except Exception:
                raise WrongParameterError(
                    "ComponentCreator Invalid --last parameter\n")
            aargs = generateDeviceNames(self.options.device, first, last,
                                        self.options.minimal)
            if self.options.datasource:
                dsargs = generateDeviceNames(
                    self.options.datasource, first, last,
                    self.options.minimal) or None
            else:
                dsargs = None
        else:
            dsargs = None

        self.args += aargs
        if not len(self.args):
            raise WrongParameterError("")

        if dsargs is None and self.options.datasource:
            dsargs = [self.options.datasource]
        if dsargs is not None and len(self.args) != len(dsargs):
            raise WrongParameterError(
                "component names cannot be match into datasource namse")
        if self.options.database:
            if not self.options.overwrite:
                existing = self._areComponentsAvailable(
                    self.args, self.options.server)
                if existing:
                    raise CPExistsException(
                        "Components '%s' already exist." % existing)
        elif not self.options.overwrite:
            existing = self._componentFilesExist(
                self.args, self.options.file, self.options.directory)
            if existing:
                raise CPExistsException(
                    "Component files '%s' already exist." % existing)

        for i, name in enumerate(self.args):
            dsname = dsargs[i] if dsargs is not None else None
            if not self.options.database:
                if self._printouts:
                    print("CREATING: %s%s.xml" % (self.options.file, name))
            else:
                if self._printouts:
                    print("STORING: %s" % (name))
            self._createComponent(
                name, self.options.directory,
                self.options.file,
                self.options.nexuspath,
                self.options.strategy,
                self.options.type,
                self.options.units,
                int(self.options.fieldlinks) + 2 * int(
                    self.options.sourcelinks),
                self.options.server if self.options.database else None,
                self.options.chunk, dsname,
                self.options.canfail
            )


class TangoDSCreator(Creator):

    """ tango datasource creator
    """

    def create(self):
        """ creates a tango datasource xml and stores it in DB or filesytem
        """
        dvargs = []
        dsargs = []
        if self.options.device.strip():
            try:
                last = int(self.options.last)
                try:
                    first = int(self.options.first)
                except Exception:
                    first = 1
                dvargs = generateDeviceNames(
                    self.options.device, first, last)
                dsargs = generateDeviceNames(
                    self.options.datasource, first, last)
            except Exception:
                dvargs = [str(self.options.device)]
                dsargs = [str(self.options.datasource)]

        if not dvargs or not len(dvargs):
            raise WrongParameterError("")

        if self.options.database:
            if not self.options.overwrite:
                existing = self._areDataSourcesAvailable(
                    dsargs, self.options.server)
                if existing:
                    raise DSExistsException(
                        "DataSources '%s' already exist." % existing)
        elif not self.options.overwrite:
            existing = self._dataSourceFilesExist(
                dsargs, self.options.file, self.options.directory)
            if existing:
                raise DSExistsException(
                    "DataSource files '%s' already exist." % existing)

        for i in range(len(dvargs)):
            if not self.options.database:
                print("CREATING %s: %s%s.ds.xml" % (
                    dvargs[i], self.options.file, dsargs[i]))
            else:
                print("STORING %s: %s" % (dvargs[i], dsargs[i]))
            self._createTangoDataSource(
                dsargs[i], self.options.directory, self.options.file,
                self.options.server if self.options.database else None,
                dvargs[i],
                self.options.attribute,
                self.options.host,
                self.options.port,
                self.options.group or None,
                self.options.elementtype or "attribute"
            )


class ClientDSCreator(Creator):

    """ client datasource creator
    """

    def create(self):
        """ creates a client datasource xml and stores it in DB or filesytem
        """
        dsargs = None
        aargs = []
        if self.options.device.strip():
            try:
                first = int(self.options.first)
            except Exception:
                raise WrongParameterError(
                    "ClientDSCreator: Invalid --first parameter\n")
            try:
                last = int(self.options.last)
            except Exception:
                raise WrongParameterError(
                    "ClientDSCreator: Invalid --last parameter\n")

            aargs = generateDeviceNames(self.options.device, first, last,
                                        self.options.minimal)
            if self.options.dsource:
                dsaargs = generateDeviceNames(
                    self.options.dsource, first, last)
                dsargs = list(self.args) + dsaargs

        self.args += aargs
        if not dsargs:
            dsargs = self.args
        if not len(self.args):
            raise WrongParameterError("")

        if self.options.database:
            if not self.options.overwrite:
                existing = self._areDataSourcesAvailable(
                    dsargs, self.options.server)
                if existing:
                    raise DSExistsException(
                        "DataSources '%s' already exist." % existing)
        elif not self.options.overwrite:
            existing = self._dataSourceFilesExist(
                dsargs, self.options.file, self.options.directory)
            if existing:
                raise DSExistsException(
                    "DataSource files '%s' already exist." % existing)

        for i in range(len(self.args)):
            if not self.options.database:
                print("CREATING: %s%s.ds.xml" % (
                    self.options.file, dsargs[i]))
            else:
                print("STORING: %s" % (dsargs[i]))
            self._createClientDataSource(
                self.args[i], self.options.directory,
                self.options.file,
                self.options.server if self.options.database else None,
                dsargs[i])


class DeviceDSCreator(Creator):

    """ device datasource creator
    """

    def create(self):
        """ creates a tango datasources xml of given device
            and stores it in DB or filesytem
        """
        if self.options.database:
            if not self.options.overwrite:
                dsargs = [
                    "%s%s" % (self.options.datasource.lower(), at.lower())
                    for at in self.args
                ]
                existing = self._areDataSourcesAvailable(
                    dsargs, self.options.server)
                if existing:
                    raise DSExistsException(
                        "DataSources '%s' already exist." % existing)
        elif not self.options.overwrite:
            dsargs = [
                "%s%s" % (self.options.datasource.lower(), at.lower())
                for at in self.args
            ]
            existing = self._dataSourceFilesExist(
                dsargs, self.options.file, self.options.directory)
            if existing:
                raise DSExistsException(
                    "DataSource files '%s' already exist." % existing)

        for at in self.args:
            dsname = "%s%s" % (self.options.datasource.lower(), at.lower())
            if not self.options.database:
                if self._printouts:
                    print("CREATING %s/%s: %s%s.ds.xml" % (
                        self.options.device, at, self.options.file, dsname))
            else:
                if self._printouts:
                    print("STORING %s/%s: %s" % (
                        self.options.device, at, dsname))
            self._createTangoDataSource(
                dsname, self.options.directory, self.options.file,
                self.options.server if self.options.database else None,
                self.options.device, at, self.options.host,
                self.options.port,
                self.options.datasource
                if not self.options.nogroup else None)


class PoolDSCreator(Creator):

    """ datasource creator of all sardana pool acquisition channels
    """

    def __init__(self, options, args, printouts=True):
        """ constructor

        :param options: command options
        :type options: :class:`optparse.Values`
        :param args: command arguments
        :type args: :obj:`list` <:obj:`str` >
        :param printouts: if printout is enable
        :type printouts: :obj:`bool`
        """
        Creator.__init__(self, options, args, printouts)
        #: (:obj:`dict` <:obj:`str`, :obj:`str` >) datasource xml dictionary
        self.datasources = {}

    def _printAction(self, dv, dscps=None):
        """ prints out information about the performed action

        :param dv: online device object
        :type dv: :class:`Device`
        :param dscps: datasource components
        :type dscps: :obj:`dict` <:obj:`str`, :obj:`list` < :obj:`str` > >
        """
        if self._printouts:
            if hasattr(self.options, "directory") and \
               self.options.directory:
                print("CREATING %s: %s/%s%s.ds.xml" % (
                    dv.tdevice, self.options.directory,
                    self.options.file, dv.name))
            elif self.options.database:
                print("CREATING %s %s/%s/%s %s" % (
                    str(dv.name) + ":" + " " * (34 - len(dv.name or "")),
                    dv.hostname,
                    dv.tdevice,
                    str(dv.attribute) + " " * (
                        70 - len(dv.tdevice or "") - len(dv.attribute or "")
                        - len(dv.hostname or "")),
                    ",".join(dscps[dv.name])
                    if (dscps and dv.name in dscps and dscps[dv.name])
                    else ""))
            else:
                print("TEST %s %s/%s/%s %s" % (
                    str(dv.name) + ":" + " " * (34 - len(dv.name or "")),
                    dv.hostname,
                    dv.tdevice,
                    str(dv.attribute) + " " * (
                        70 - len(dv.tdevice or "") - len(dv.attribute or "")
                        - len(dv.hostname or "")),
                    ",".join(dscps[dv.name])
                    if (dscps and dv.name in dscps and dscps[dv.name])
                    else ""))

    def create(self):
        """ creates datasources of all online.xml simple devices
        """
        self.createXMLs()
        server = self.options.server
        if not hasattr(self.options, "directory") or \
           not self.options.directory:
            if self.options.database:
                for dsname, dsxml in self.datasources.items():
                    storeDataSource(dsname, dsxml, server)
        else:
            for dsname, dsxml in self.datasources.items():
                with open("%s/%s%s.ds.xml" % (
                        self.options.directory,
                        self.options.file, dsname), "w") as myfile:
                    myfile.write(dsxml)

    def __createDevice(self, name, source, clientlike=True):
        """  create Device from name source and chtype

        :param name: alias device name
        :type name: :obj:`str`
        :param source: device source string
        :type source: :obj:`str`
        :param clientlike: device to be client like
        :type clientlike: :obj:`bool`
        :returns: device object
        :rtype: :obj:`Device`
        """
        dv = Device()
        if name and source:
            if source.startswith("tango://"):
                source = source[8:]
            slst = source.split('/')
            if not slst[0] or ":" not in slst[0] or len(slst) < 2:
                return dv
            hplst = slst[0].split(":")
            if len(hplst) != 2:
                return dv
            dv.host, dv.port = hplst[0], hplst[1]
            dv.hostname = slst[0]
            dv.attribute = slst[-1]
            dv.name = name
            dv.tdevice = "/".join(slst[1:-1])
            if clientlike:
                dv.group = "__CLIENT__"
        return dv

    def createXMLs(self):
        """ creates datasource xmls of all online.xml simple devices
        """
        self.datasources = {}
        plname = self.options.pool
        motors = ['Motor', 'PseudoMotor']
        expchnls = ['CTExpChannel', 'ZeroDExpChannel',
                    'OneDExpChannel', 'TwoDExpChannel',
                    'PseudoCounter']
        if 'ALL' in self.args or not self.args:
            args = list(motors)
            args.extend(expchnls)
        else:
            args = list(self.args)
        try:
            plproxy = openServer(plname)
        except Exception:
            raise Exception("Cannot connect to %s" % plname)
        try:
            chlist = plproxy.ExpChannelList
        except Exception:
            chlist = []
        try:
            motorlist = plproxy.MotorList
        except Exception:
            motorlist = []
        for ellist in [chlist, motorlist]:
            for els in ellist:
                elprop = json.loads(els) or {}
                if 'name' in elprop.keys() \
                   and 'source' in elprop.keys() \
                   and 'type' in elprop.keys():
                    if elprop['name'] in args or elprop['type'] in args:
                        dv = self.__createDevice(
                            elprop['name'], elprop['source'],
                            self.options.clientlike)
                        if self.options.lower:
                            dv.tolower()
                        if dv.tdevice and dv.attribute and dv.host and dv.port:
                            self._printAction(dv)
                            xml = self._createTangoDataSource(
                                dv.name, None, None, None,
                                dv.tdevice, dv.attribute, dv.host,
                                dv.port, dv.group)
                            self.datasources[dv.name] = xml


class OnlineDSCreator(Creator):

    """ datasource creator of all online.xml simple devices
    """

    def __init__(self, options, args, printouts=True):
        """ constructor

        :param options: command options
        :type options: :class:`optparse.Values`
        :param args: command arguments
        :type args: :obj:`list` <:obj:`str` >
        :param printouts: if printout is enable
        :type printouts: :obj:`bool`
        """
        Creator.__init__(self, options, args, printouts)
        #: (:obj:`dict` <:obj:`str`, :obj:`str` >) datasource xml dictionary
        self.datasources = {}
        if options.xmlpackage:
            xmlPackageHandler.loadXMLTemplates(options.xmlpackage)
        else:
            xmlPackageHandler.loadXMLTemplates('nxstools.xmltemplates')
        #: (:obj:`str`) xml template component package path
        self.xmltemplatepath = xmlPackageHandler.packagepath
        #: (:obj:`str`) xml template component package
        self.xmlpackage = xmlPackageHandler.package

    def _printAction(self, dv, dscps=None):
        """ prints out information about the performed action

        :param dv: online device object
        :type dv: :class:`Device`
        :param dscps: datasource components
        :type dscps: :obj:`dict` <:obj:`str`, :obj:`list` < :obj:`str` > >
        """
        if self._printouts:
            if hasattr(self.options, "directory") and \
               self.options.directory:
                print("CREATING %s: %s/%s%s.ds.xml" % (
                    dv.tdevice, self.options.directory,
                    self.options.file, dv.name))
            elif self.options.database:
                print("CREATING %s %s/%s/%s %s" % (
                    str(dv.name) + ":" + " " * (34 - len(dv.name or "")),
                    dv.hostname,
                    dv.tdevice,
                    str(dv.attribute) + " " * (
                        70 - len(dv.tdevice or "") - len(dv.attribute or "")
                        - len(dv.hostname or "")),
                    ",".join(dscps[dv.name])
                    if (dscps and dv.name in dscps and dscps[dv.name])
                    else ""))
            else:
                print("TEST %s %s %s %s/%s/%s %s" % (
                    str(dv.name) + ":" + " " * (34 - len(dv.name or "")),
                    str(dv.dtype) + ":" + " " * (20 - len(dv.dtype or "")),
                    str(dv.module) + ":" + " " * (24 - len(dv.module or "")),
                    dv.hostname,
                    dv.tdevice,
                    str(dv.attribute) + " " * (
                        70 - len(dv.tdevice or "") - len(dv.attribute or "")
                        - len(dv.hostname or "")),
                    ",".join(dscps[dv.name])
                    if (dscps and dv.name in dscps and dscps[dv.name])
                    else ""))

    def create(self):
        """ creates datasources of all online.xml simple devices
        """
        self.createXMLs()
        server = self.options.server
        if not hasattr(self.options, "directory") or \
           not self.options.directory:
            if self.options.database:
                for dsname, dsxml in self.datasources.items():
                    storeDataSource(dsname, dsxml, server)
        else:
            for dsname, dsxml in self.datasources.items():
                with open("%s/%s%s.ds.xml" % (
                        self.options.directory,
                        self.options.file, dsname), "w") as myfile:
                    myfile.write(dsxml)

    def createXMLs(self):
        """ creates datasource xmls of all online.xml simple devices
        """
        self.datasources = {}
        tangohost = getServerTangoHost(
            self.options.external or self.options.server)
        hw = etree.parse(self.args[0],
                         parser=XMLParser(collect_ids=False)).getroot()
        if hw.tag != 'hw':
            hw = hw.find('hw')
        dscps = {}
        if self.options.server and self._printouts and \
           (
               not hasattr(self.options, "directory") or
               not self.options.directory
           ):
            try:
                dscps = getDataSourceComponents(
                    self.options.server, self.options.verbose)
            except Exception:
                dscps = {}

        for device in hw:
            if device.tag == 'device':
                dv = Device()
                dv.name = self._getChildText(device, "name")
                dv.dtype = self._getChildText(device, "type")
                dv.module = self._getChildText(device, "module")
                dv.tdevice = self._getChildText(device, "device")
                dv.hostname = self._getChildText(device, "hostname")
                dv.sardananame = self._getChildText(device, "sardananame")
                dv.sardanahostname = self._getChildText(
                    device, "sardanahostname")
                if self.options.lower:
                    dv.tolower()
                try:
                    dv.splitHostPort()
                except Exception:
                    if self._printouts:
                        print("ERROR %s: host for module %s of %s "
                              "type not defined"
                              % (dv.name, dv.module, dv.dtype))
                    continue
                dv.findAttribute(tangohost, self.options.clientlike)
                created = False
                if dv.attribute:
                    dv.setSardanaName(self.options.lower)
                    mdv = copy.copy(dv)
                    mdv.tdevice = dv.sdevice or dv.tdevice
                    self._printAction(mdv, dscps)
                    xml = self._createTangoDataSource(
                        mdv.name, None, None, None,
                        mdv.tdevice, mdv.attribute, mdv.host,
                        mdv.port, mdv.group)
                    self.datasources[mdv.name] = xml
                    created = True
                module = self._getModuleName(dv)
                smodule = "%s@pool" % module.lower() if module else None
                if module and module.lower() in \
                   self.xmlpackage.moduleMultiAttributes.keys():
                    multattr = self.xmlpackage.moduleMultiAttributes[
                        module.lower()]
                    for at in multattr:
                        dsname = "%s_%s" % (dv.name.lower(), at.lower())
                        xml = self._createTangoDataSource(
                            dsname, None, None, None,
                            dv.tdevice, at, dv.thost, dv.tport,
                            "%s_" % (dv.name))
                        self.datasources[dsname] = xml
                        mdv = copy.copy(dv)
                        mdv.name = dsname
                        mdv.hostname = "%s:%s" % (dv.thost, dv.tport)
                        mdv.attribute = at
                        self._printAction(mdv, dscps)
                    created = True
                if smodule in \
                   self.xmlpackage.moduleMultiAttributes.keys():
                    smultattr = self.xmlpackage.moduleMultiAttributes[
                        smodule]
                    if smultattr and not dv.sdevice:
                        if self._printouts:
                            print(
                                "SKIPPING %s: Device cannot be found" %
                                dv.name)
                    else:
                        for at in smultattr:
                            dsname = "%s_%s" % (
                                dv.name, at.lower())
                            xml = self._createTangoDataSource(
                                dsname, None, None, None,
                                dv.sdevice, at, dv.shost, dv.sport,
                                "%s_" % (dv.name))
                            #   "__CLIENT__")
                            self.datasources[dsname] = xml
                            mdv = copy.copy(dv)
                            mdv.name = dsname
                            mdv.tdevice = dv.sdevice
                            mdv.hostname = "%s:%s" % (dv.shost, dv.sport)
                            mdv.attribute = at
                            self._printAction(mdv, dscps)
                        created = True
                if not created:
                    if self._printouts:
                        print(
                            "SKIPPING %s:    module '%s' of '%s' "
                            "type not defined"
                            % (dv.name, dv.module, dv.dtype))


class CPCreator(Creator):

    """ component creator of all online.xml complex devices
    """

    def __init__(self, options, args, printouts=True):
        """ constructor

        :param options: command options
        :type options: :class:`optparse.Values`
        :param args: command arguments
        :type args: :obj:`list` <:obj:`str` >
        :param printouts: if printout is enable
        :type printouts: :obj:`bool`
        """
        Creator.__init__(self, options, args, printouts)
        #: (:obj:`dict` <:obj:`str`, :obj:`str` >) datasource xml dictionary
        self.datasources = {}
        #: (:obj:`dict` <:obj:`str`, :obj:`str` >) component xml dictionary
        self.components = {}
        #: component xml dictionary
        if options.xmlpackage:
            xmlPackageHandler.loadXMLTemplates(options.xmlpackage)
        else:
            xmlPackageHandler.loadXMLTemplates('nxstools.xmltemplates')
        #: (:obj:`str`) xml template component package path
        self.xmltemplatepath = xmlPackageHandler.packagepath
        #: (:obj:`str`) xml template component package
        self.xmlpackage = xmlPackageHandler.package

    def create(self):
        """ creates components of all online.xml complex devices
        """
        cpname = self.options.component
        if hasattr(self.options, "database") and \
           self.options.database:
            server = self.options.server
            if not self.options.overwrite:
                if self._areComponentsAvailable(
                        [cpname], server, self.options.lower):
                    raise CPExistsException(
                        "Component '%s' already exists." % cpname)
        elif not self.options.overwrite:
            existing = self._componentFilesExist(
                [cpname], self.options.file, self.options.directory)
            if existing:
                raise CPExistsException(
                    "Component files '%s' already exist." % existing)

        self.createXMLs()
        if not self.datasources and not self.components:
            raise CPExistsException(
                "Warning: Component %s cannot be created" % cpname)
        server = self.options.server
        if hasattr(self.options, "database") and \
           self.options.database:
            for dsname, dsxml in self.datasources.items():
                storeDataSource(dsname, dsxml, server)
            for cpname, cpxml in self.components.items():
                mand = False
                if hasattr(self.options, "mandatory") and \
                   self.options.mandatory:
                    mand = True
                storeComponent(cpname, cpxml, server, mand)
        else:
            for dsname, dsxml in self.datasources.items():
                with open("%s/%s%s.ds.xml" % (
                        self.options.directory,
                        self.options.file, dsname), "w") as myfile:
                    myfile.write(dsxml)
            for cpname, cpxml in self.components.items():
                with open("%s/%s%s.xml" % (
                        self.options.directory,
                        self.options.file, cpname), "w") as myfile:
                    myfile.write(cpxml)

    @classmethod
    def _replaceName(cls, filename, cpname, module=None):
        """ replaces name prefix of xml templates files

        :param filename: template filename
        :type filename: :obj:`str`
        :param cpname: output prefix
        :type cpname: :obj:`str`
        :param module: module name
        :type module: :obj:`str`
        :returns: output filename
        :rtype: :obj:`str`
        """
        if filename.endswith(".ds.xml"):
            filename = filename[:-7]
        elif filename.endswith(".xml"):
            filename = filename[:-4]
        sname = filename.split("_")
        if not module or module == sname[0]:
            sname[0] = cpname
        return "_".join(sname)

    def createXMLs(self):
        """ creates component xmls of all online.xml complex devices
        abstract method
        """
        pass


class CompareOnlineDS(object):

    """ comparing tool for online.xml files
    """

    def __init__(self, options, args, printouts=True):
        """ constructor

        :param options:  command options
        :type options: :class:`optparse.Values`
        :param args: command arguments
        :type args: :obj:`list` < :obj:`str` >
        :param printouts: if printout is enable
        :type printouts: :obj:`bool`
        """
        #: (:class:`optparse.Values`) creator options
        self.options = options
        #: (:obj:`list` < :obj:`str` >) creator arguments
        self.args = args
        #: (:obj:`bool`) if printout is enable
        self._printouts = printouts

    @classmethod
    def _getText(cls, node):
        """ provides xml content of the node

        :param node: DOM node
        :type node: :class:`lxml.etree.Element`
        :returns: xml content string
        :rtype: :obj:`str`
        """
        if node is None:
            return
        xml = _toxml(node)
        start = xml.find('>')
        end = xml.rfind('<')
        if start == -1 or end < start:
            return ""
        return xml[start + 1:end].replace("&lt;", "<").replace("&gt;", "<"). \
            replace("&quot;", "\"").replace("&amp;", "&")

    @classmethod
    def _getChildText(cls, parent, childname):
        """ provides text of child named by childname

        :param parent: parent node
        :type parent: :class:`lxml.etree.Element`
        :param childname: child name
        :type childname: :opj:`str`
        :returns: text string
        :rtype: :obj:`str`
        """

        children = parent.findall(childname)
        return cls._getText(children[0]) if len(children) else None

    def _load(self, fname):
        """ loads device data from online.xml file

        :param fname: filename
        :type fname: :obj:`str`
        :returns: dictionary with devices of the given name
        :rtype: :obj:`dict` <:obj:`str`, :obj:`list` <:class:`Device`>>
        """

        dct = {}
        hw = etree.parse(fname,
                         parser=XMLParser(collect_ids=False)).getroot()
        if hw.tag != 'hw':
            hw = hw.find('hw')
        for device in hw:
            if device.tag == 'device':
                name = self._getChildText(device, "name")
                if self.options.lower:
                    name = name.lower()
                dv = Device()
                dv.name = name
                dv.dtype = self._getChildText(device, "type")
                dv.module = self._getChildText(device, "module")
                dv.tdevice = self._getChildText(device, "device")
                dv.hostname = self._getChildText(device, "hostname")
                dv.sardananame = self._getChildText(device, "sardananame")
                dv.sardanahostname = self._getChildText(
                    device, "sardanahostname")
                sname = dv.sardananame or name
                if sname not in dct.keys():
                    dct[sname] = []
                dct[sname].append(dv)
        return dct

    def compare(self):
        if self._printouts:
            print("Comparing: %s\n" % " ".join(self.args))
        dct1 = self._load(self.args[0])
        dct2 = self._load(self.args[1])
        common = sorted(set(dct1.keys()) & set(dct2.keys()))
        d1md2 = sorted(set(dct1.keys()) - set(dct2.keys()))
        d2md1 = sorted(set(dct2.keys()) - set(dct1.keys()))
        addd1 = dict((str(k),
                      [(str(dv.name) if dv.name else dv.name)
                       for dv in dct1[k]])
                     for k in d1md2)
        addd2 = dict((str(k),
                      [(str(dv.name) if dv.name else dv.name)
                       for dv in dct2[k]])
                     for k in d2md1)
        diff = {}
        for name in common:
            ndiff = {}
            l1 = [True] * len(dct1[name])
            l2 = [True] * len(dct2[name])
            for i1, dv1 in enumerate(dct1[name]):
                for i2, dv2 in enumerate(dct2[name]):
                    if l1[i1] and l2[i2]:
                        res = dv1.compare(dv2)
                        if not res:
                            l1[i1] = False
                            l2[i2] = False
                            break
                        else:
                            ndiff["%s:%s" % (i1, i2)] = res
            if True in l1 and True not in l2:
                addd1[str(name)] = []
                for i1, dv in enumerate(dct1[name]):
                    if l1[i1]:
                        addd1[str(name)].append(
                            (str(dv.name) if dv.name else dv.name))
            elif True not in l1 and True in l2:
                addd2[str(name)] = []
                for i2, dv in enumerate(dct2[name]):
                    if l2[i2]:
                        addd2[str(name)].append(
                            (str(dv.name) if dv.name else dv.name))
            if True in l1 or True in l2:
                diff[str(name)] = []
                for i1, dv1 in enumerate(dct1[name]):
                    for i2, dv2 in enumerate(dct2[name]):
                        if l1[i1] and l2[i2]:
                            diff[str(name)].append(ndiff["%s:%s" % (i1, i2)])

        if self._printouts:
            import pprint
            print("Additional devices in '%s' {alias: [name]} :\n"
                  % self.args[0])
            pprint.pprint(addd1)
            print("\nAdditional devices in '%s' {alias: [name]} :\n"
                  % self.args[1])
            pprint.pprint(addd2)
            print("\nDiffrences in the common part:\n")
            pprint.pprint(diff)


class OnlineCPCreator(CPCreator):

    """ component creator of online components
    """

    def __init__(self, options, args, printouts=True):
        """ constructor

        :param options: command options
        :type options: :class:`optparse.Values`
        :param args: command arguments
        :type args: :obj:`list` < :obj:`str` >
        :param printouts: if printout is enable
        :type printouts: :obj:`bool`
        """
        CPCreator.__init__(self, options, args, printouts)

    def _printAction(self, dv, dscps=None):
        """ prints out information about the performed action

        :param dv: online device object
        :type dv: :class:`Device`
        :param dscps: datasource components
        :type dscps: :obj:`dict` <:obj:`str`, :obj:`list` < :obj:`str` > >
        """
        if self._printouts:
            if hasattr(self.options, "database") and \
               self.options.database:
                print("CREATING %s %s/%s %s" % (
                    str(dv.name) + ":" + " " * (34 - len(dv.name)),
                    dv.hostname,
                    str(dv.tdevice) + " " * (
                        60 - len(dv.tdevice) - len(dv.hostname)),
                    ",".join(dscps[dv.name])
                    if (dscps and dv.name in dscps and dscps[dv.name])
                    else ""))
            else:
                print("CREATING %s: %s/%s%s.ds.xml" % (
                    dv.tdevice, self.options.directory, self.options.file,
                    dv.name))

    def listcomponents(self):
        """ provides a list of components with xml templates

        :returns: list of components with xml templates
        :rtype: :obj:`list` <:obj:`str` >
        """
        hw = etree.parse(
            self.args[0],
            parser=XMLParser(collect_ids=False)).getroot()
        if hw.tag != 'hw':
            hw = hw.find('hw')
        cpnames = set()
        for device in hw:
            if device.tag == 'device':
                dvname = self._getChildText(device, "name")
                sardananame = self._getChildText(device, "sardananame")
                name = sardananame or dvname
                if self.options.lower:
                    name = name.lower()
                dv = Device()
                dv.name = name
                dv.dtype = self._getChildText(device, "type")
                dv.module = self._getChildText(device, "module")
                dv.tdevice = self._getChildText(device, "device")
                dv.hostname = self._getChildText(device, "hostname")
                dv.sardananame = self._getChildText(device, "sardananame")
                dv.sardanahostname = self._getChildText(
                    device, "sardanahostname")

                module = self._getModuleName(dv)
                if module:
                    if module.lower() in self.xmlpackage.moduleTemplateFiles:
                        cpnames.add(dv.name)
        return cpnames

    def listcomponenttypes(self):
        """ provides a list of standard component types

        :returns: list of standard component types
        :rtype: :obj:`list` <:obj:`str`>
        """
        return list(sorted(self.xmlpackage.moduleTemplateFiles.keys()))

    def createXMLs(self):
        """ creates component xmls of all online.xml complex devices
        """
        self.datasources = {}
        self.components = {}
        if self.options.component and self.options.cptype and \
           self.options.device:
            hw = [None]
        else:
            hw = etree.parse(
                self.args[0],
                parser=XMLParser(collect_ids=False)).getroot()
            if hw.tag != 'hw':
                hw = hw.find('hw')
        cpname = self.options.component
        tangohost = getServerTangoHost(
            self.options.external or self.options.server)
        for device in hw:
            if device is None or device.tag == 'device':
                if device is None:
                    name = cpname
                else:
                    dvname = self._getChildText(device, "name")
                    sardananame = self._getChildText(
                        device, "sardananame")
                    name = sardananame or dvname
                if self.options.lower:
                    name = name.lower()
                    cpname = cpname.lower()
                if name == cpname:
                    dv = Device()
                    dv.name = name
                    dv.dtype = self.options.cptype or \
                        self._getChildText(device, "type")
                    dv.module = self.options.cptype or \
                        self._getChildText(device, "module")
                    dv.tdevice = self.options.device or \
                        self._getChildText(device, "device")
                    hostname = None
                    if self.options.host:
                        host = self.options.host
                        port = self.options.port \
                            if self.options.port else "10000"
                        hostname = "%s:%s" % (host, port)
                    if hostname is None and device is None:
                        hostname = tangohost
                    if hostname is not None:
                        dv.hostname = hostname
                    if hostname is None and device is not None:
                        dv.hostname = self._getChildText(device, "hostname")
                    if device is not None:
                        dv.sardananame = \
                            self._getChildText(device, "sardananame")
                        dv.sardanahostname = self._getChildText(
                            device, "sardanahostname")

                    dv.findDevice(tangohost)
                    try:
                        dv.splitHostPort()
                    except Exception:
                        if self._printouts:
                            print("ERROR %s: host for module %s of %s "
                                  "type not defined"
                                  % (dv.name, dv.module, dv.dtype))
                        continue
                    if self.options.cptype:
                        module = self.options.cptype
                    else:
                        module = self._getModuleName(dv)
                    if module:
                        if module.lower() in \
                           self.xmlpackage.moduleMultiAttributes.keys():
                            multattr = self.xmlpackage.moduleMultiAttributes[
                                module.lower()]
                            for at in multattr:
                                dsname = "%s_%s" % (
                                    dv.name, at.lower())
                                xml = self._createTangoDataSource(
                                    dsname, None, None, None,
                                    dv.tdevice, at, dv.host, dv.port,
                                    "%s_" % (dv.name.lower()))
                                self.datasources[dsname] = xml
                                mdv = copy.copy(dv)
                                mdv.name = dsname
                                mdv.attribute = at
                                self._printAction(mdv)
                        smodule = "%s@pool" % module.lower()
                        if smodule in \
                           self.xmlpackage.moduleMultiAttributes.keys():
                            smultattr = self.xmlpackage.moduleMultiAttributes[
                                smodule]
                            if smultattr and not dv.sdevice:
                                raise Exception(
                                    "Device %s cannot be found" % dv.name)
                            for at in smultattr:
                                dsname = "%s_%s" % (
                                    dv.name,
                                    at.lower())
                                xml = self._createTangoDataSource(
                                    dsname, None, None, None,
                                    dv.sdevice, at, dv.shost, dv.sport,
                                    "%s_" % (dv.name))
                                #   "__CLIENT__")
                                self.datasources[dsname] = xml
                                mdv = copy.copy(dv)
                                mdv.name = dsname
                                mdv.tdevice = dv.sdevice
                                mdv.hostname = "%s:%s" % (dv.shost, dv.sport)
                                mdv.attribute = at
                                self._printAction(mdv)
                        if module.lower() \
                           in self.xmlpackage.moduleTemplateFiles:
                            xmlfiles = self.xmlpackage.moduleTemplateFiles[
                                module.lower()]
                            for xmlfile in xmlfiles:
                                newname = self._replaceName(xmlfile, cpname)
                                with open(
                                        '%s/%s' % (
                                            self.xmltemplatepath, xmlfile), "r"
                                ) as content_file:
                                    xmlcontent = content_file.read()
                                xml = xmlcontent.replace("$(name)", cpname)\
                                    .replace("$(device)", dv.tdevice)\
                                    .replace("$(__entryname__)",
                                             (self.options.entryname
                                              or "scan"))\
                                    .replace("$(__insname__)",
                                             (self.options.insname
                                              or "instrument"))\
                                    .replace("$(hostname)", dv.hostname)
                                mdv = copy.copy(dv)
                                mdv.name = newname
                                self._printAction(mdv)
                                if xmlfile.endswith(".ds.xml"):
                                    self.datasources[newname] = xml
                                else:
                                    self.components[newname] = xml


class StandardCPCreator(CPCreator):

    """ component creator of standard templates
    """

    def __init__(self, options, args, printouts=True):
        """ constructor

        :param options: command options
        :type options: :class:`optparse.Values`
        :param args: command arguments
        :type args: :obj:`list` < :obj:`str` >
        :param printouts: if printout is enable
        :type printouts: :obj:`bool`
        """
        CPCreator.__init__(self, options, args, printouts)
        self.__params = {}
        self.__specialparams = {}

    def listcomponenttypes(self):
        """ provides a list of standard component types

        :returns: list of standard component types
        :rtype: :obj:`list` <:obj:`str`>
        """
        return list(sorted(self.xmlpackage.standardComponentVariables.keys()))

    def listcomponentvariables(self):
        """ provides a list of standard component types

        :returns: list of standard component types
        :rtype: :obj:`list` <:obj:`str`>
        """

        if self.options.cptype not \
           in self.xmlpackage.standardComponentVariables.keys():
            raise Exception(
                "Component type %s not in %s" %
                (self.options.cptype,
                 list(self.xmlpackage.standardComponentVariables.keys())))
        return self.xmlpackage.standardComponentVariables[
            self.options.cptype]

    def __setspecialparams(self):
        """ sets special parameters,
        i.e. __tangohost__, __tangoport__ and __configdevice__

        """
        server = self.options.external or self.options.server
        host, port = getServerTangoHost(server).split(":")
        self.__specialparams['__tangohost__'] = host
        self.__specialparams['__tangoport__'] = port
        if server:
            proxy = openServer(server)
            self.__specialparams['__configdevice__'] = proxy.name()
        else:
            self.__specialparams['__configdevice__'] = None

    def createXMLs(self):
        """ creates component xmls of all online.xml complex devices
        """
        self.datasources = {}
        self.components = {}
        self.__setspecialparams()
        if self.args:
            self.__params = dict(zip(self.args[::2], self.args[1::2]))
        else:
            self.__params = {}
        cpname = self.options.component
        module = self.options.cptype
        if self.options.lower:
            cpname = cpname.lower()
            module = module.lower()
        if module not in self.xmlpackage.standardComponentVariables.keys():
            raise Exception(
                "Component type %s not in %s" %
                (module,
                 list(self.xmlpackage.standardComponentVariables.keys()))
            )

        xmlfiles = []
        if module in self.xmlpackage.standardComponentTemplateFiles:
            xmlfiles = self.xmlpackage.standardComponentTemplateFiles[module]
        else:
            if os.path.isfile("%s/%s.xml" % (self.xmltemplatepath, module)):
                xmlfiles = ["%s.xml" % module]
        for xmlfile in xmlfiles:
            print(xmlfile)
            newname = self._replaceName(xmlfile, cpname, module)
            with open(
                    '%s/%s' % (
                        self.xmltemplatepath, xmlfile), "r"
            ) as content_file:
                xmlcontent = content_file.read()
                xml = xmlcontent.replace("$(name)", cpname).replace(
                    "$(__entryname__)",
                    (self.options.entryname or "scan")).replace(
                        "$(__insname__)",
                        (self.options.insname
                         or "instrument"))
                missing = []
                for var, desc in self.xmlpackage.standardComponentVariables[
                        module].items():
                    if var in self.__params.keys():
                        xml = xml.replace("$(%s)" % var, self.__params[var])
                    elif var in self.__specialparams.keys():
                        if self.__specialparams[var] is not None:
                            xml = xml.replace("$(%s)" % var,
                                              self.__specialparams[var])
                        else:
                            raise Exception(
                                "Parameter: %s cannot be found" % var)
                    elif desc["default"] is not None:
                        xml = xml.replace("$(%s)" % var, desc["default"])
                    else:
                        missing.append(var)
                if missing:
                    if sys.version_info > (3,):
                        root = et.fromstring(
                            bytes(xml, "UTF-8"),
                            parser=XMLParser(collect_ids=False))
                    else:
                        root = et.fromstring(
                            xml,
                            parser=XMLParser(collect_ids=False))
                    nodes = root.findall(".//attribute")
                    nodes.extend(root.findall(".//field"))
                    nodes.extend(root.findall(".//link"))
                    grnodes = root.findall(".//group")
                    for node in nodes:
                        text = self.__getText(node)
                        for ms in missing:
                            label = "$(%s)" % ms
                            if label in text:
                                parent = node.getparent()
                                parent.remove(node)
                                break
                    for node in grnodes:
                        text = node.attrib["name"]
                        if text and "$(" in text:
                            for ms in missing:
                                label = "$(%s)" % ms
                                if label in text:
                                    parent = node.getparent()
                                    parent.remove(node)
                                    break
                    xml = _simpletoxml(root)
                    if self._printouts:
                        print("MISSING %s" % missing)
                    errors = []
                    for var in missing:
                        if "s.$(%s)" % var in xml:
                            errors.append(var)
                    if errors:
                        print(
                            "WARNING: %s cannot be created without %s"
                            % (var, errors))
                        continue

                    for var in missing:
                        xml = xml.replace("$(%s)" % var, "")
                    lines = xml.split('\n')
                    xml = '\n'.join([x for x in lines if len(x.strip())])
                if xmlfile.endswith(".ds.xml"):
                    self._printAction(newname)
                    self.datasources[newname] = xml
                else:
                    self._printAction(newname)
                    self.components[newname] = xml

    def _printAction(self, name):
        """ prints out information about the performed action

        :param name: component name
        :type name: :obj:`str`
        """
        if self._printouts:
            if hasattr(self.options, "database") and \
               self.options.database:
                print("CREATING '%s' of '%s' on '%s' with %s" % (
                    name,
                    self.options.cptype,
                    self.options.server,
                    self.__params))
            else:
                print("CREATING '%s' of '%s' in '%s/%s%s.xml' with %s" % (
                    name,
                    self.options.cptype,
                    self.options.directory,
                    self.options.file,
                    name,
                    self.__params))

    @classmethod
    def __getText(cls, node):
        """ collects text from text child nodes

        :param node: parent node
        :type node: :obj:`xml.etree.ElementTree.Element`
        """
        if node is not None:
            tnodes = ([node.text] if node.text else []) \
                     + [child.tail for child in node if child.tail]
            return unicode("".join(tnodes)).strip()
        return ""


if __name__ == "__main__":
    pass



#!/usr/bin/env python
#   This file is part of nexdatas - Tango Server for NeXus data writer
#
#    Copyright (C) 2012-2018 DESY, Jan Kotanski <jkotan@mail.desy.de>
#
#    nexdatas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    nexdatas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with nexdatas.  If not, see <http://www.gnu.org/licenses/>.
#

"""  Creator of XML configuration files """


import PyTango
import sys

import lxml.etree


if sys.version_info > (3,):
    unicode = str


def _tostr(text):
    """ convert bytestr or unicode to python str
    :param text: text to convert
    :type text: :obj:`bytes` or :obj:`unicode` or :obj:`str`
    :returns: converted text
    :rtype: :obj:`str`
    """
    if isinstance(text, str):
        return text
    else:
        if sys.version_info > (3,) and \
           (isinstance(text, bytes) or isinstance(text, unicode)):
            return str(text, "utf8")
        else:
            return str(text)


class NTag(object):

    """ tag wrapper
    """

    #

    def __init__(self, parent, tagName, nameAttr="", typeAttr=""):
        """ constructor

        :param parent: parent tag element
        :type parent: :class:`NTag`
        :param tagName: tag name
        :type tagName: :obj:`str`
        :param nameAttr: value of name attribute
        :type nameAttr: :obj:`str`
        :param typeAttr: value of type attribute
        :type typeAttr: :obj:`str`
        """

        #: (:class:`lxml.etree.Element`) tag element from etree
        self.elem = lxml.etree.Element(tagName)
        parent.elem.append(self.elem)

        if nameAttr != "":
            self.elem.attrib["name"] = nameAttr
        if typeAttr != "":
            self.elem.attrib["type"] = typeAttr

    def addTagAttr(self, name, value):
        """ adds tag attribute

        :param name: attribute name
        :type name: :obj:`str`
        :param value: attribute value
        :type value: :obj:`str`
        """
        self.elem.attrib[name] = value

    def setText(self, text):
        """ sets tag content

        :param text: tag content
        :type text: :obj:`str`
        """
        self.elem.text = text

    def addText(self, text):
        """ adds tag content

        :param text: tag content
        :type text: :obj:`str`
        """
        self.elem.text = self.elem.text + text


class NAttr(NTag):

    """ Attribute tag wrapper
    """

    def __init__(self, parent, nameAttr, typeAttr=""):
        """ constructor

        :param parent: parent tag element
        :type parent: :class:`NTag`
        :param nameAttr: name attribute
        :type nameAttr: :obj:`str`
        :param typeAttr: type attribute
        :type typeAttr: :obj:`str`
        """
        NTag.__init__(self, parent, "attribute", nameAttr, typeAttr)

    def setStrategy(self, mode="STEP", trigger=None, value=None, canfail=None):
        """ sets the attribute strategy

        :param mode: mode data writing, i.e. INIT, STEP, FINAL, POSTRUN
        :type mode: :obj:`str`
        :param trigger: for asynchronous writting,
                        e.g. with different subentries
        :type trigger: :obj:`str`
        :param value: label for postrun mode
        :type value: :obj:`str`
        :param canfail: can fail strategy flag
        :type canfail: :obj:`bool`
        """
        #: strategy of data writing, i.e. INIT, STEP, FINAL, POSTRUN
        strategy = NTag(self, "strategy")
        if strategy:
            strategy.addTagAttr("mode", mode)
        if trigger:
            strategy.addTagAttr("trigger", trigger)
        if canfail:
            strategy.addTagAttr("canfail", "true")
        if value:
            strategy.setText(value)


class NGroup(NTag):

    """ Group tag wrapper
    """

    def __init__(self, parent, nameAttr, typeAttr=""):
        """ constructor

        :param parent: parent tag element
        :type parent: :class:`NTag`
        :param nameAttr: name attribute
        :type nameAttr: :obj:`str`
        :param typeAttr: type attribute
        :type typeAttr: :obj:`str`
        """
        NTag.__init__(self, parent, "group", nameAttr, typeAttr)
        #: (:obj:`list` <:obj:`str`> ) list of doc tag contents
        self._doc = []
        #: (:obj:`dict` <:obj:`str` , :obj:`str`> )
        #:    container with attribute tag wrappers
        self._gAttr = {}

    def addDoc(self, doc):
        """ adds doc tag content

        :param doc: doc tag content
        :type doc: :obj:`str`
        """
        self._doc.append(NTag(self, "doc"))
        self._doc[-1].addText(doc)

    def addAttr(self, attrName, attrType, attrValue=""):
        """adds attribute: tag

        :param attrName: name attribute
        :type attrName: :obj:`str`
        :param attrType: type attribute
        :type attrType: :obj:`str`
        :param attrValue: content of the attribute tag
        :type attrValue: :obj:`str`
        """
        print("%s %s %s" % (attrName, attrType, attrValue))
        at = NAttr(self, attrName, attrType)
        self._gAttr[attrName] = at
        if attrValue != "":
            at.setText(attrValue)
        return self._gAttr[attrName]


class NLink(NTag):

    """ Link tag wrapper
    """

    def __init__(self, parent, nameAttr, gTarget):
        """ constructor

        :param parent: parent tag element
        :type parent: :class:`NTag`
        :param nameAttr: name attribute
        :type nameAttr: :obj:`str`
        :param gTarget: target attribute
        :type gTarget: :obj:`str`
        """
        NTag.__init__(self, parent, "link", nameAttr)
        self.addTagAttr("target", gTarget)
        #: list of doc tag contents
        self._doc = []

    def addDoc(self, doc):
        """ adds doc tag content

        :param doc: doc tag content
        :type doc: :obj:`str`
        """
        self._doc.append(NTag(self, "doc"))
        self._doc[-1].addText(doc)


class NDimensions(NTag):

    """ Dimensions tag wrapper
    """

    def __init__(self, parent, rankAttr):
        """ constructor

        :param parent: parent tag element
        :type parent: :class:`NTag`
        :param rankAttr: rank attribute
        :type rankAttr: :obj:`str`
        """
        NTag.__init__(self, parent, "dimensions")
        self.addTagAttr("rank", rankAttr)
        #: (:obj:`dict` <:obj:`str`, :class:`NDim`>)
        #:    container with dim tag wrapper
        self.dims = {}

    def dim(self, indexAttr, valueAttr):
        """ adds dim tag

        :param indexAttr: index attribute
        :type indexAttr: :obj:`str`
        :param valueAttr: value attribute
        :type valueAttr: :obj:`str`
        """
        self.dims[indexAttr] = NDim(self, indexAttr, valueAttr)


class NDim(NTag):

    """ Dim tag wrapper
    """

    def __init__(self, parent, indexAttr, valueAttr):
        """ constructor

        :param parent: parent tag element
        :type parent: :class:`NTag`
        :param indexAttr: index attribute
        :type indexAttr: :obj:`str`
        :param valueAttr: value attribute
        :type valueAttr: :obj:`str`
        """
        NTag.__init__(self, parent, "dim")
        self.addTagAttr("index", indexAttr)
        self.addTagAttr("value", valueAttr)


class NField(NTag):

    """ Field tag wrapper
    """

    def __init__(self, parent, nameAttr, typeAttr=""):
        """constructor

        :param parent: parent tag element
        :type parent: :class:`NTag`
        :param nameAttr: name attribute
        :type nameAttr: :obj:`str`
        :param typeAttr: type attribute
        :type typeAttr: :obj:`str`
        """
        NTag.__init__(self, parent, "field", nameAttr, typeAttr)

        #: (:obj:`list` <:obj:`str`> ) list of doc tag contents
        self._doc = []
        #: (:obj:`dict` <:obj:`str` , :obj:`str`> )
        #:    container with attribute tag wrappers
        self._attr = {}

    def setStrategy(self, mode="STEP", trigger=None, value=None,
                    grows=None, compression=False, rate=None,
                    shuffle=None, canfail=None, compression_opts=None):
        """ sets the field strategy

        :param mode: mode data writing, i.e. INIT, STEP, FINAL, POSTRUN
        :type mode: :obj:`str`
        :param trigger: for asynchronous writting,
                        e.g. with different subentries
        :type trigger: :obj:`str`
        :param value: label for postrun mode
        :type value: :obj:`str`
        :param grows: growing dimension
        :type grows: :obj:`str`
        :param compression: flag if compression shuold be applied
        :type compression: :obj:`str`
        :param rate: compression rate
        :type rate: :obj:`str`
        :param shuffle: flag if compression shuffle
        :type shuffle: :obj:`str`
        :param canfail: can fail strategy flag
        :type canfail: :obj:`bool`
        """
        #: strategy of data writing, i.e. INIT, STEP, FINAL, POSTRUN
        strategy = NTag(self, "strategy")
        if strategy:
            strategy.addTagAttr("mode", mode)
        if grows:
            strategy.addTagAttr("grows", grows)
        if trigger:
            strategy.addTagAttr("trigger", trigger)
        if value:
            strategy.setText(value)
        if canfail:
            strategy.addTagAttr("canfail", "true")
        if compression:
            if int(compression) == 1:
                strategy.addTagAttr("compression", "true")
                if rate is not None:
                    strategy.addTagAttr("rate", str(rate))
            else:
                strategy.addTagAttr(
                    "compression", int(compression))
                if compression_opts:
                    strategy.addTagAttr(
                        "compression_opts",
                        ",".join([str(opts) for opts in compression_opts]))
            if shuffle is not None:
                strategy.addTagAttr(
                    "shuffle",
                    "true" if shuffle else "false")

    def setUnits(self, unitsAttr):
        """ sets the field unit

        :param unitsAttr: the field unit
        :type unitsAttr: :obj:`str`
        """
        self.addTagAttr("units", unitsAttr)

    def addDoc(self, doc):
        """ adds doc tag content

        :param doc: doc tag content
        :type doc: :obj:`str`
        """
        self._doc.append(NTag(self, "doc"))
        self._doc[-1].addText(doc)

    def addAttr(self, attrName, attrType, attrValue=""):
        """ adds attribute tag

        :param attrName: name attribute
        :type attrName: :obj:`str`
        :param attrType: type attribute
        :type attrType: :obj:`str`
        :param attrValue: content of the attribute tag
        :type attrValue: :obj:`str`
        """
        self._attr[attrName] = NAttr(self, attrName, attrType)
        if attrValue != '':
            self._attr[attrName].setText(attrValue)
        return self._attr[attrName]


class NDSource(NTag):

    """ Source tag wrapper
    """

    def __init__(self, parent):
        """ constructor

        :param parent: parent tag element
        :type parent: :class:`NTag`
        """
        NTag.__init__(self, parent, "datasource")

        #: list of doc tag contents
        self._doc = []

    def initDBase(self, name, dbtype, query, dbname=None, rank=None,
                  mycnf=None, user=None,
                  passwd=None, dsn=None, mode=None, host=None, port=None):
        """ sets parameters of DataBase

        :param name: name of datasource
        :type name: :obj:`str`
        :param dbname: name of used DataBase
        :type dbname: :obj:`str`
        :param query: database query
        :type query: :obj:`str`
        :param dbtype: type of the database, i.e. MYSQL, PGSQL, ORACLE
        :type dbtype: :obj:`str`
        :param rank: rank of the query output, i.e. SCALAR, SPECTRUM, IMAGE
        :type rank: :obj:`str`
        :param mycnf: MYSQL config file
        :type mycnf: :obj:`str`
        :param user: database user name
        :type user: :obj:`str`
        :param passwd: database user password
        :type passwd: :obj:`str`
        :param dsn: DSN string to initialize ORACLE and PGSQL databases
        :type dsn: :obj:`str`
        :param mode: mode for ORACLE databases, i.e. SYSDBA or SYSOPER
        :type mode: :obj:`str`
        :param host: name of the host
        :type host: :obj:`str`
        :param port: port number
        :type port: :obj:`str`
        """
        self.addTagAttr("type", "DB")
        self.addTagAttr("name", name)
        da = NTag(self, "database")
        da.addTagAttr("dbtype", dbtype)

        if host:
            da.addTagAttr("hostname", host)
        if port:
            da.addTagAttr("port", port)
        if dbname:
            da.addTagAttr("dbname", dbname)
        if user:
            da.addTagAttr("user", user)
        if passwd:
            da.addTagAttr("passwd", passwd)
        if mycnf:
            da.addTagAttr("mycnf", mycnf)
        if mode:
            da.addTagAttr("mode", mode)
        if dsn:
            da.addText(dsn)

        da = NTag(self, "query")
        if rank:
            da.addTagAttr("format", rank)
        da.addText(query)

    def initTango(self, name, device, memberType, recordName, host=None,
                  port=None, encoding=None, group=None):
        """ sets paramters for Tango device

        :param name: name of datasource
        :type name: :obj:`str`
        :param device: device name
        :type device: :obj:`str`
        :param memberType: type of the data object, i.e. attribute,
                           property, command
        :type memberType: :obj:`str`
        :param recordName: name of the data object
        :type recordName: :obj:`str`
        :param host: host name
        :type host: :obj:`str`
        :param port: port
        :type port: :obj:`str`
        :param encoding: encoding of DevEncoded data
        :type encoding: :obj:`str`
        """
        self.addTagAttr("type", "TANGO")
        self.addTagAttr("name", name)
        dv = NTag(self, "device")
        dv.addTagAttr("name", device)

        if memberType:
            dv.addTagAttr("member", memberType)
        if host:
            dv.addTagAttr("hostname", host)
        if port:
            dv.addTagAttr("port", port)
        if encoding:
            dv.addTagAttr("encoding", encoding)
        if group:
            dv.addTagAttr("group", group)

        da = NTag(self, "record")
        da.addTagAttr("name", recordName)

    def initClient(self, name, recordName):
        """ sets paramters for Client data

        :param name: name of datasource
        :type name: :obj:`str`
        :param recordName: name of the data object
        :type recordName: :obj:`str`
        """
        self.addTagAttr("type", "CLIENT")
        self.addTagAttr("name", name)
        da = NTag(self, "record")
        da.addTagAttr("name", recordName)

    def addDoc(self, doc):
        """ adds doc tag content

        :param doc: doc tag content
        :type doc: :obj:`str`
        """
        self._doc.append(NTag(self, "doc"))
        self._doc[-1].addText(doc)


class NDeviceGroup(NGroup):

    """ Tango device tag creator
    """

    #: (:obj:`list` < :obj:`str`>) Tango types
    tTypes = ["DevVoid",
              "DevBoolean",
              "DevShort",
              "DevLong",
              "DevFloat",
              "DevDouble",
              "DevUShort",
              "DevULong",
              "DevString",
              "DevVarCharArray",
              "DevVarShortArray",
              "DevVarLongArray",
              "DevVarFloatArray",
              "DevVarDoubleArray",
              "DevVarUShortArray",
              "DevVarULongArray",
              "DevVarStringArray",
              "DevVarLongStringArray",
              "DevVarDoubleStringArray",
              "DevState",
              "ConstDevString",
              "DevVarBooleanArray",
              "DevUChar",
              "DevLong64",
              "DevULong64",
              "DevVarLong64Array",
              "DevVarULong64Array",
              "DevInt",
              "DevEncoded"]

    #: (:obj:`list` <:obj:`str`>) NeXuS types corresponding to the Tango types
    nTypes = ["NX_CHAR",
              "NX_BOOLEAN",
              "NX_INT32",
              "NX_INT32",
              "NX_FLOAT32",
              "NX_FLOAT64",
              "NX_UINT32",
              "NX_UINT32",
              "NX_CHAR",
              "NX_CHAR",
              "NX_INT32",
              "NX_INT32",
              "NX_FLOAT32",
              "NX_FLOAT64",
              "NX_UINT32",
              "NX_UINT32",
              "NX_CHAR",
              "NX_CHAR",
              "NX_CHAR",
              "NX_CHAR",
              "NX_CHAR",
              "NX_BOOLEAN",
              "NX_CHAR",
              "NX_INT64",
              "NX_UINT64",
              "NX_INT64",
              "NX_UINT64",
              "NX_INT32",
              "NX_CHAR"]

    def __init__(self, parent, deviceName, nameAttr, typeAttr="",
                 commands=True, blackAttrs=None):
        """ constructor

        :param parent: parent tag element
        :type parent: :class:`NTag`
        :param deviceName: tango device name
        :type deviceName: :obj:`str`
        :param nameAttr: name attribute
        :type nameAttr: :obj:`str`
        :param typeAttr: type attribute
        :type typeAttr: :obj:`str`
        :param commands: if we call the commands
        :type commands: :obj:`bool`
        :param blackAttrs: list of excluded attributes
        :type blackAttrs: :obj:`list` <:obj:`str`>
        """
        NGroup.__init__(self, parent, nameAttr, typeAttr)
        #: (:class:`PyTango.DeviceProxy`) device proxy
        self._proxy = PyTango.DeviceProxy(deviceName)
        #: (:obj:`dict` <:obj:`str`, :class:`NTag`>) fields of the device
        self._fields = {}
        #: (:obj:`list` <:obj:`str`>) blacklist for Attributes
        self._blackAttrs = blackAttrs if blackAttrs else []
        #: (:obj:`str`) the device name
        self._deviceName = deviceName

        self._fetchProperties()
        self._fetchAttributes()
        if commands:
            self._fetchCommands()

    def _fetchProperties(self):
        """ fetches properties

        :brief: It collects the device properties
        """
        prop = self._proxy.get_property_list('*')
        print("PROPERIES %s" % prop)
        for pr in prop:
            self.addAttr(pr, "NX_CHAR",
                         str(self._proxy.get_property(pr)[pr][0]))
            if pr not in self._fields:
                self._fields[pr] = NField(self, pr, "NX_CHAR")
                self._fields[pr].setStrategy("STEP")
                sr = NDSource(self._fields[pr])
                sr.initTango(
                    self._deviceName, self._deviceName, "property",
                    pr, host="haso228k.desy.de", port="10000")

    def _fetchAttributes(self):
        """ fetches Attributes

        :brief: collects the device attributes
        """

        #: device attirbutes
        attr = self._proxy.get_attribute_list()
        for at in attr:

            print(at)
            cf = self._proxy.attribute_query(at)
            print("QUERY")
            print(cf)
            print(cf.name)
            print(cf.data_format)
            print(cf.standard_unit)
            print(cf.display_unit)
            print(cf.unit)
            print(self.tTypes[cf.data_type])
            print(self.nTypes[cf.data_type])
            print(cf.data_type)

            if at not in self._fields and at not in self._blackAttrs:
                self._fields[at] = NField(self, at, self.nTypes[cf.data_type])
                encoding = None
                if str(cf.data_format).split('.')[-1] == "SPECTRUM":
                    da = self._proxy.read_attribute(at)
                    d = NDimensions(self._fields[at], "1")
                    d.dim("1", str(da.dim_x))
                    if str(da.type) == 'DevEncoded':
                        encoding = 'VDEO'
                if str(cf.data_format).split('.')[-1] == "IMAGE":
                    da = self._proxy.read_attribute(at)
                    d = NDimensions(self._fields[at], "2")
                    d.dim("1", str(da.dim_x))
                    d.dim("2", str(da.dim_y))
                    if str(da.type) == 'DevEncoded':
                        encoding = 'VDEO'

                if cf.unit != 'No unit':
                    self._fields[at].setUnits(cf.unit)
                    self._fields[at].setUnits(cf.unit)

                if cf.description != 'No description':
                    self._fields[at].addDoc(cf.description)
                self.addAttr('URL', "NX_CHAR", "tango://" + self._deviceName)

                self._fields[at].setStrategy("STEP")
                sr = NDSource(self._fields[at])
                sr.initTango(self._deviceName, self._deviceName, "attribute",
                             at, host="haso228k.desy.de", port="10000",
                             encoding=encoding)

    def _fetchCommands(self):
        """ fetches commands

        :brief: It collects results of the device commands
        """
        #: list of the device commands
        cmd = self._proxy.command_list_query()
        print("COMMANDS %s" % cmd)
        for cd in cmd:
            if str(cd.in_type).split(".")[-1] == "DevVoid" \
                    and str(cd.out_type).split(".")[-1] != "DevVoid" \
                    and str(cd.out_type).split(".")[-1] in self.tTypes \
                    and cd.cmd_name not in self._fields:
                self._fields[cd.cmd_name] = \
                    NField(
                        self, cd.cmd_name,
                        self.nTypes[self.tTypes.index(
                            str(cd.out_type).split(".")[-1])])
                self._fields[cd.cmd_name].setStrategy("STEP")
                sr = NDSource(self._fields[cd.cmd_name])
                sr.initTango(self._deviceName, self._deviceName,
                             "command", cd.cmd_name,
                             host="haso228k.desy.de", port="10000")


class XMLFile(object):

    """ XML file object
    """

    def __init__(self, fname):
        """ constructor

        :param fname: XML file name
        :type fname: :obj:`str`
        """
        #: (:obj:`str`) XML file name
        self.fname = fname
        #: (:class:`lxml.etree.Element`) XML root instance
        self.elem = lxml.etree.Element("definition")

    def prettyPrint(self, etNode=None):
        """prints pretty XML making use of etree

        :param etNode: node
        :type etNode: :class:`lxml.etree.Element`
        """
        node = etNode if etNode is not None else self.elem
        xmls = _tostr(
            lxml.etree.tostring(
                node, encoding='utf8',
                method='xml', pretty_print=True))
        if not xmls.startswith("<?xml"):
            xmls = "<?xml version='1.0' encoding='utf8'?>\n" + xmls
        return xmls

    def dump(self):
        """ dumps XML structure into the XML file

        :brief: It opens XML file, calls prettyPrint and closes the XML file
        """
        with open(self.fname, "w") as myfile:
            myfile.write(self.prettyPrint(self.elem))


def main():
    """ the main function
    """
    #: handler to XML file
    df = XMLFile("test.xml")
    #: entry
    en = NGroup(df, "entry1", "NXentry")
    #: instrument
    ins = NGroup(en, "instrument", "NXinstrument")
    #:    NXsource
    src = NGroup(ins, "source", "NXsource")
    #: field
    f = NField(src, "distance", "NX_FLOAT")
    f.setUnits("m")
    f.setText("100.")

    df.dump()


if __name__ == "__main__":
    main()
